"""
Borrowed from
https://github.com/asyncdef/apyio/blob/master/apyio/__init__.py
"""

"""Async compatibility wrappers for Python io module stream objects."""

import io as sync_io


def open(
        file,
        mode="r",
        buffering=-1,
        encoding=None,
        errors=None,
        newline=None,
        closefd=True,
        opener=None,
):
    r"""Open file and return a stream.  Raise OSError upon failure.

    file is either a text or byte string giving the name (and the path
    if the file isn't in the current working directory) of the file to
    be opened or an integer file descriptor of the file to be
    wrapped. (If a file descriptor is given, it is closed when the
    returned I/O object is closed, unless closefd is set to False.)

    mode is an optional string that specifies the mode in which the file is
    opened. It defaults to 'r' which means open for reading in text mode. Other
    common values are 'w' for writing (truncating the file if it already
    exists), 'x' for exclusive creation of a new file, and 'a' for appending
    (which on some Unix systems, means that all writes append to the end of the
    file regardless of the current seek position). In text mode, if encoding is
    not specified the encoding used is platform dependent. (For reading and
    writing raw bytes use binary mode and leave encoding unspecified.) The
    available modes are:

    ========= ===============================================================
    Character Meaning
    --------- ---------------------------------------------------------------
    'r'       open for reading (default)
    'w'       open for writing, truncating the file first
    'x'       create a new file and open it for writing
    'a'       open for writing, appending to the end of the file if it exists
    'b'       binary mode
    't'       text mode (default)
    '+'       open a disk file for updating (reading and writing)
    'U'       universal newline mode (deprecated)
    ========= ===============================================================

    The default mode is 'rt' (open for reading text). For binary random
    access, the mode 'w+b' opens and truncates the file to 0 bytes, while
    'r+b' opens the file without truncation. The 'x' mode implies 'w' and
    raises an `FileExistsError` if the file already exists.

    Python distinguishes between files opened in binary and text modes,
    even when the underlying operating system doesn't. Files opened in
    binary mode (appending 'b' to the mode argument) return contents as
    bytes objects without any decoding. In text mode (the default, or when
    't' is appended to the mode argument), the contents of the file are
    returned as strings, the bytes having been first decoded using a
    platform-dependent encoding or using the specified encoding if given.

    'U' mode is deprecated and will raise an exception in future versions
    of Python.  It has no effect in Python 3.  Use newline to control
    universal newlines mode.

    buffering is an optional integer used to set the buffering policy.
    Pass 0 to switch buffering off (only allowed in binary mode), 1 to select
    line buffering (only usable in text mode), and an integer > 1 to indicate
    the size of a fixed-size chunk buffer.  When no buffering argument is
    given, the default buffering policy works as follows:

    * Binary files are buffered in fixed-size chunks; the size of the buffer
      is chosen using a heuristic trying to determine the underlying device's
      "block size" and falling back on `io.DEFAULT_BUFFER_SIZE`.
      On many systems, the buffer will typically be 4096 or 8192 bytes long.

    * "Interactive" text files (files for which isatty() returns True)
      use line buffering.  Other text files use the policy described above
      for binary files.

    encoding is the str name of the encoding used to decode or encode the
    file. This should only be used in text mode. The default encoding is
    platform dependent, but any encoding supported by Python can be
    passed.  See the codecs module for the list of supported encodings.

    errors is an optional string that specifies how encoding errors are to
    be handled---this argument should not be used in binary mode. Pass
    'strict' to raise a ValueError exception if there is an encoding error
    (the default of None has the same effect), or pass 'ignore' to ignore
    errors. (Note that ignoring encoding errors can lead to data loss.)
    See the documentation for codecs.register for a list of the permitted
    encoding error strings.

    newline is a string controlling how universal newlines works (it only
    applies to text mode). It can be None, '', '\n', '\r', and '\r\n'.  It
    works as follows:

    * On input, if newline is None, universal newlines mode is
      enabled. Lines in the input can end in '\n', '\r', or '\r\n', and
      these are translated into '\n' before being returned to the
      caller. If it is '', universal newline mode is enabled, but line
      endings are returned to the caller untranslated. If it has any of
      the other legal values, input lines are only terminated by the given
      string, and the line ending is returned to the caller untranslated.

    * On output, if newline is None, any '\n' characters written are
      translated to the system default line separator, os.linesep. If
      newline is '', no translation takes place. If newline is any of the
      other legal values, any '\n' characters written are translated to
      the given string.

    closedfd is a bool. If closefd is False, the underlying file descriptor
    will be kept open when the file is closed. This does not work when a file
    name is given and must be True in that case.

    The newly created file is non-inheritable.

    A custom opener can be used by passing a callable as *opener*. The
    underlying file descriptor for the file object is then obtained by calling
    *opener* with (*file*, *flags*). *opener* must return an open file
    descriptor (passing os.open as *opener* results in functionality similar to
    passing None).

    open() returns a file object whose type depends on the mode, and
    through which the standard file operations such as reading and writing
    are performed. When open() is used to open a file in a text mode ('w',
    'r', 'wt', 'rt', etc.), it returns a TextIOWrapper. When used to open
    a file in a binary mode, the returned class varies: in read binary
    mode, it returns a BufferedReader; in write binary and append binary
    modes, it returns a BufferedWriter, and in read/write mode, it returns
    a BufferedRandom.

    It is also possible to use a string or bytearray as a file for both
    reading and writing. For strings StringIO can be used like a file
    opened in a text mode, and for bytes a BytesIO can be used like a file
    opened in a binary mode.
    """
    result = sync_io.open(
        file,
        mode,
        buffering,
        encoding,
        errors,
        newline,
        closefd,
        opener,
    )

    return wrap_file(result)


def wrap_file(file_like_obj):
    """Wrap a file like object in an async stream wrapper.

    Files generated with `open()` may be one of several types. This
    convenience function retruns the stream wrapped in the most appropriate
    wrapper for the type. If the stream is already wrapped it is returned
    unaltered.
    """
    if isinstance(file_like_obj, AsyncIOBaseWrapper):

        return file_like_obj

    if isinstance(file_like_obj, sync_io.FileIO):

        return AsyncFileIOWrapper(file_like_obj)

    if isinstance(file_like_obj, sync_io.BufferedRandom):

        return AsyncBufferedRandomWrapper(file_like_obj)

    if isinstance(file_like_obj, sync_io.BufferedReader):

        return AsyncBufferedReaderWrapper(file_like_obj)

    if isinstance(file_like_obj, sync_io.BufferedWriter):

        return AsyncBufferedWriterWrapper(file_like_obj)

    if isinstance(file_like_obj, sync_io.TextIOWrapper):

        return AsyncTextIOWrapperWrapper(file_like_obj)

    raise TypeError(
        'Unrecognized file stream type {}.'.format(file_like_obj.__class__),
    )


class AsyncIOBaseWrapper:

    """IOBase object wrapper for async compatibility."""

    def __init__(self, stream):
        """Wrap an IOBase stream object."""
        self._stream = stream

    async def seek(self, pos, whence=0):
        """Change stream position.

        Change the stream position to byte offset pos. Argument pos is
        interpreted relative to the position indicated by whence.  Values
        for whence are ints:

        * 0 -- start of stream (the default); offset should be zero or positive
        * 1 -- current stream position; offset may be negative
        * 2 -- end of stream; offset is usually negative
        Some operating systems / file systems could provide additional values.

        Return an int indicating the new absolute position.
        """
        return self._stream.seek(pos, whence)

    async def tell(self):
        """Return an int indicating the current stream position."""
        return self._stream.tell()

    async def truncate(self, pos=None):
        """Truncate file to size bytes.

        Size defaults to the current IO position as reported by tell().  Return
        the new size.
        """
        return self._stream.truncate(pos)

    async def flush(self):
        """Flush write buffers, if applicable.

        This is not implemented for read-only and non-blocking streams.
        """
        return self._stream.flush()

    async def drain(self):
        """Flush the write buffers, if applicable.

        This exists for compatibility with the async stream drain method which
        forces a buffer flush.
        """
        return self._stream.flush()

    def close(self):
        """Flush and close the IO object.

        This method has no effect if the file is already closed.
        """
        return self._stream.close()

    def seekable(self):
        """Return a bool indicating whether object supports random access.

        If False, seek(), tell() and truncate() raise UnsupportedOperation.
        This method may need to do a test seek().
        """
        return self._stream.seekable()

    def readable(self):
        """Return a bool indicating whether object was opened for reading.

        If False, read() will raise UnsupportedOperation.
        """
        return self._stream.readable()

    def writable(self):
        """Return a bool indicating whether object was opened for writing.

        If False, write() and truncate() will raise UnsupportedOperation.
        """
        return self._stream.writable()

    @property
    def closed(self):
        """closed: bool.  True if the file has been closed.

        For backwards compatibility, this is a property, not a predicate.
        """
        return self._stream.closed

    async def __aenter__(self):
        """Context management protocol.  Returns self."""
        self._stream.__enter__()
        return self

    async def __aexit__(self, *args):
        """Context management protocol.  Calls close()."""
        self._stream.__exit__(*args)

    def fileno(self):
        """Return the underlying file descriptor (an int) if one exists.

        An OSError is raised if the IO object does not use a file descriptor.
        """
        return self._stream.fileno()

    def isatty(self):
        """Return a bool indicating whether this is an 'interactive' stream.

        Return False if it can't be determined.
        """
        return self._stream.isatty()

    async def readline(self, size=-1):
        r"""Read and return a line of bytes from the stream.

        If size is specified, at most size bytes will be read.
        Size should be an int.

        The line terminator is always b'\n' for binary files; for text
        files, the newlines argument to open can be used to select the line
        terminator(s) recognized.
        """
        return self._stream.readline(size)

    async def __aiter__(self):
        """Return self."""
        return self

    async def __anext__(self):
        """Get the next value. Convert StopIteration to StopAsyncIteration."""
        try:

            return self._stream.__next__()

        except StopIteration as exc:

            raise StopAsyncIteration() from exc

    async def readlines(self, hint=None):
        """Return all the lines."""
        return self._stream.readlines(hint)

    def writelines(self, lines):
        """Write an iterable of lines to the stream."""
        return self._stream.writelines(lines)

sync_io.IOBase.register(AsyncIOBaseWrapper)


class AsyncRawIOBaseWrapper(AsyncIOBaseWrapper):

    """RawIOBase object wrapper for async compatibility."""

    async def read(self, size=-1):
        """Read and return up to size bytes, where size is an int.

        Returns an empty bytes object on EOF, or None if the object is
        set not to block and has no data to read.
        """
        return self._stream.read(size)

    async def readall(self):
        """Read until EOF, using multiple read() call."""
        return self._stream.readall()

    async def readinto(self, b):
        """Read up to len(b) bytes into bytearray b.

        Returns an int representing the number of bytes read (0 for EOF), or
        None if the object is set not to block and has no data to read.
        """
        return self._stream.readinto(b)

    def write(self, b):
        """Write the given buffer to the IO stream.

        Returns the number of bytes written, which may be less than len(b).
        """
        return self._stream.write(b)

sync_io.RawIOBase.register(AsyncRawIOBaseWrapper)


class AsyncBufferedIOBaseWrapper(AsyncIOBaseWrapper):

    """BufferedIOBase object wrapper for async compatibility."""

    async def read(self, size=None):
        """Read and return up to size bytes, where size is an int.

        If the argument is omitted, None, or negative, reads and
        returns all data until EOF.

        If the argument is positive, and the underlying raw stream is
        not 'interactive', multiple raw reads may be issued to satisfy
        the byte count (unless EOF is reached first).  But for
        interactive raw streams (XXX and for pipes?), at most one raw
        read will be issued, and a short result does not imply that
        EOF is imminent.

        Returns an empty bytes array on EOF.

        Raises BlockingIOError if the underlying raw stream has no
        data at the moment.
        """
        return self._stream.read(size)

    async def read1(self, size=None):
        """Read up to size bytes with at most one read() system call."""
        return self._stream.read1(size)

    async def readinto(self, b):
        """Read up to len(b) bytes into bytearray b.

        Like read(), this may issue multiple reads to the underlying raw
        stream, unless the latter is 'interactive'.

        Returns an int representing the number of bytes read (0 for EOF).

        Raises BlockingIOError if the underlying raw stream has no
        data at the moment.
        """
        return self._stream.readinto(b)

    async def readinto1(self, b):
        """Read up to len(b) bytes into *b*, using at most one system call.

        Returns an int representing the number of bytes read (0 for EOF).

        Raises BlockingIOError if the underlying raw stream has no
        data at the moment.
        """
        return self._stream.readinto1(b)

    def write(self, b):
        """Write the given bytes buffer to the IO stream.

        Return the number of bytes written, which is never less than
        len(b).

        Raises BlockingIOError if the buffer is full and the
        underlying raw stream cannot accept more data at the moment.
        """
        return self._stream.write(b)

    def detach(self):
        """
        Separate the underlying raw stream from the buffer and return it.

        After the raw stream has been detached, the buffer is in an unusable
        state.
        """
        return self._stream.detach()

sync_io.BufferedIOBase.register(AsyncBufferedIOBaseWrapper)


class AsyncBytesIOWrapper(AsyncBufferedIOBaseWrapper):

    """BytesIO object wrapper for async compatibility."""

    def __getstate__(self):
        """Return a pickle dictionary."""
        body = self.__dict__.copy()
        body['_stream'] = body['_stream'].__getstate__()
        return body

    def getvalue(self):
        """Return the bytes value (contents) of the buffer."""
        return self._stream.getvalue()

    def getbuffer(self):
        """Return a readable and writable view of the buffer."""
        return self._stream.getbuffer()

    def close(self):
        """Close the stream."""
        return self._stream.close()

    async def read(self, size=None):
        """Read from the stream."""
        return self._stream.read(size)

    async def read1(self, size):
        """Read from the stream."""
        return self._stream.read1(size)

    def write(self, b):
        """Write to the stream."""
        return self._stream.write(b)

    async def seek(self, pos, whence=0):
        """Change the stream position."""
        return self._stream.seek(pos, whence)

    async def tell(self):
        """Get the stream position."""
        return self._stream.tell()

    async def truncate(self, pos=None):
        """Truncate the stream."""
        return self._stream.truncate(pos)

    def readable(self):
        """Get if the stream is readable."""
        return self._stream.readable()

    def writable(self):
        """Get if the stream is writable."""
        return self._stream.writable()

    def seekable(self):
        """Get if the stream is seekable."""
        return self._stream.seekable()


class AsyncBufferedReaderWrapper(AsyncBufferedIOBaseWrapper):

    """BufferedReader object wrapper for async compatibility."""

    async def read(self, size=None):
        """Read size bytes.

        Returns exactly size bytes of data unless the underlying raw IO
        stream reaches EOF or if the call would block in non-blocking
        mode. If size is negative, read until EOF or until read() would
        block.
        """
        return self._stream.read(size)

    async def peek(self, size=0):
        """Return buffered bytes without advancing the position.

        The argument indicates a desired minimal number of bytes; we
        do at most one raw read to satisfy it.  We never return more
        than self.buffer_size.
        """
        return self._stream.peek(size)

    async def read1(self, size):
        """Read up to size bytes, with at most one read() system call."""
        return self._stream.read1(size)

    async def tell(self):
        """Get the current position."""
        return self._stream.tell()

    async def seek(self, pos, whence=0):
        """Change the stream position."""
        return self._stream.seek(pos, whence)


class AsyncBufferedWriterWrapper(AsyncBufferedIOBaseWrapper):

    """BufferedWriter object wrapper for async compatibility."""

    def write(self, b):
        """Write to the stream."""
        return self._stream.write(b)

    async def truncate(self, pos=None):
        """Truncate the stream."""
        return self._stream.truncate(pos)

    async def flush(self):
        """Flush the stream buffer."""
        return self._stream.flush()

    async def tell(self):
        """Get the current stream position."""
        return self._stream.tell()

    async def seek(self, pos, whence=0):
        """Change the stream position."""
        return self._stream.seek(pos, whence)


class AsyncBufferedRandomWrapper(
        AsyncBufferedWriterWrapper,
        AsyncBufferedReaderWrapper,
):

    """BufferedRandom object wrapper for async compatibility."""


class AsyncFileIOWrapper(AsyncRawIOBaseWrapper):

    """FileIO object wrapper for async compatibility."""

    def __getstate__(self):
        """Get the pickle dictionary for the stream."""
        body = self.__dict__.copy()
        body['_stream'] = body['_stream'].__getstate__()

    def __repr__(self):
        """Get the human representation of the wrapper."""
        return '<AsyncFileIOWrapper {}>'.format(self._stream)

    async def read(self, size=None):
        """Read at most size bytes, returned as bytes.

        Only makes one system call, so less data may be returned than requested
        In non-blocking mode, returns None if no data is available.
        Return an empty bytes object at EOF.
        """
        return self._stream.read(size)

    async def readall(self):
        """Read all data from the file, returned as bytes.

        In non-blocking mode, returns as much as is immediately available,
        or None if no data is available.  Return an empty bytes object at EOF.
        """
        return self._stream.readall()

    async def readinto(self, b):
        """Replicate the RawIO readinto."""
        return self._stream.readinto(b)

    def write(self, b):
        """Write bytes b to file, return number written.

        Only makes one system call, so not all of the data may be written.
        The number of bytes actually written is returned.  In non-blocking
        mode, returns None if the write would block.
        """
        return self._stream.write(b)

    async def seek(self, pos, whence=sync_io.SEEK_SET):
        """Move to new file position.

        Argument offset is a byte count.  Optional argument whence defaults to
        SEEK_SET or 0 (offset from start of file, offset should be >= 0); other
        values are SEEK_CUR or 1 (move relative to current position, positive
        or negative), and SEEK_END or 2 (move relative to end of file, usually
        negative, although many platforms allow seeking beyond the end of a
        file).

        Note that not all file objects are seekable.
        """
        return self._stream.seek(pos, whence)

    async def tell(self):
        """Get the current file position.

        Can raise OSError for non seekable files.
        """
        return self._stream.tell()

    async def truncate(self, size=None):
        """Truncate the file to at most size bytes.

        Size defaults to the current file position, as returned by tell().
        The current file position is changed to the value of size.
        """
        return self._stream.truncate(size)

    def close(self):
        """Close the file.

        A closed file cannot be used for further I/O operations.  close() may
        be called more than once without error.
        """
        return self._stream.close()

    def seekable(self):
        """True if file supports random-access."""
        return self._stream.seekable()

    def readable(self):
        """True if file was opened in a read mode."""
        return self._stream.readable()

    def writable(self):
        """True if file was opened in a write mode."""
        return self._stream.readable()

    def fileno(self):
        """Return the underlying file descriptor (an integer)."""
        return self._stream.fileno()

    def isatty(self):
        """True if the file is connected to a TTY device."""
        return self._stream.isatty()

    @property
    def closefd(self):
        """True if the file descriptor will be closed by close()."""
        return self._stream.closefd

    @property
    def mode(self):
        """Return the string giving the file mode."""
        return self._stream.mode


class AsyncTextIOBaseWrapper(AsyncIOBaseWrapper):

    """TextIOBase object wrapper for async compatibility."""

    async def read(self, size=-1):
        """Read at most size characters from stream, where size is an int.

        Read from underlying buffer until we have size characters or we hit
        EOF. If size is negative or omitted, read until EOF.

        Returns a string.
        """
        return self._stream.read(size)

    def write(self, s):
        """Write string s to stream and returning an int."""
        return self._stream.write(s)

    async def truncate(self, pos=None):
        """Truncate size to pos, where pos is an int."""
        return self._stream.truncate(pos)

    async def readline(self):
        """Read until newline or EOF.

        Returns an empty string if EOF is hit immediately.
        """
        return self._stream.readlines()

    def detach(self):
        """
        Separate the underlying buffer from the TextIOBase and return it.

        After the underlying buffer has been detached, the TextIO is in an
        unusable state.
        """
        return self._stream.detach()

    @property
    def encoding(self):
        """Get the stream encoding."""
        return self._stream.encoding

    @property
    def newlines(self):
        """Line endings translated so far.

        Only line endings translated during reading are considered.

        Subclasses should override.
        """
        return self._stream.newlines

    @property
    def errors(self):
        """Error setting of the decoder or encoder.

        Subclasses should override.
        """
        return self._stream.errors

sync_io.TextIOBase.register(AsyncTextIOBaseWrapper)


class AsyncTextIOWrapperWrapper(AsyncTextIOBaseWrapper):

    """TextIOWrapper object wrapper for async compatibility."""

    def __repr__(self):
        """Get the human representation of the wrapper."""
        return '<AsyncTextIOWrapperWrapper {}>'.format(self._stream)

    @property
    def encoding(self):
        """Get the stream encoding."""
        return self._stream.encoding

    @property
    def errors(self):
        """Get the stream errors."""
        return self._stream.errors

    @property
    def line_buffering(self):
        """Get the stream line buffering."""
        return self._stream.line_buffering

    @property
    def buffer(self):
        """Get the stream buffer."""
        return self._stream.buffer

    def seekable(self):
        """Get if the stream is seekable."""
        return self._stream.seekable()

    def readable(self):
        """Get if the stream is readable."""
        return self._stream.readable()

    def writable(self):
        """Get if the stream is writable."""
        return self._stream.writable()

    async def flush(self):
        """Flush the stream buffer."""
        return self._stream.flush()

    def close(self):
        """Close the stream."""
        return self._stream.close()

    @property
    def closed(self):
        """Get if the stream is closed."""
        return self._stream.closed

    @property
    def name(self):
        """Get the stream name."""
        return self._stream.name

    def fileno(self):
        """Get the stream fileno."""
        return self._stream.fileno()

    def isatty(self):
        """Get if the stream is a tty."""
        return self._stream.isatty()

    def write(self, s):
        """Write data, where s is a str."""
        return self._stream.write(s)

    async def tell(self):
        """Get the current stream position."""
        return self._stream.tell()

    async def truncate(self, pos=None):
        """Truncate the stream."""
        return self._stream.truncate(pos)

    def detach(self):
        """Detach the underlying stream buffer."""
        return self._stream.detach()

    async def seek(self, cookie, whence=0):
        """Set the stream position."""
        return self._stream.seek(cookie, whence)

    async def read(self, size=None):
        """Read from the stream."""
        return self._stream.read(size)

    async def __anext__(self):
        """Call the stream __next__."""
        try:
            return self._stream.__next__()

        except StopIteration as exc:

            raise StopAsyncIteration() from exc

    async def readline(self, size=None):
        """Read a line from the stream."""
        return self._stream.readline(size)

    @property
    def newlines(self):
        """Get the stream newlines."""
        return self._stream.newlines


class AsyncStringIOWrapper(AsyncTextIOWrapperWrapper):

    """StringIO object wrapper for async compatibility."""

    def getvalue(self):
        """Get the stream value."""
        return self._stream.getvalue()

    def __repr__(self):
        """Get the human representation of the wrapper."""
        return '<AsyncStringIOWrapper {}>'.format(self._stream)

    @property
    def errors(self):
        """Get the stream errors."""
        return self._stream.errors

    @property
    def encoding(self):
        """Get the stream encoding."""
        return self._stream.encoding

    def detach(self):
        """Detach the underlying stream buffer."""
        return self._stream.detach


def StringIO(*args, **kwargs):
    """StringIO constructor shim for the async wrapper."""
    raw = sync_io.StringIO(*args, **kwargs)
    return AsyncStringIOWrapper(raw)


def BytesIO(*args, **kwargs):
    """BytesIO constructor shim for the async wrapper."""
    raw = sync_io.BytesIO(*args, **kwargs)
    return AsyncBytesIOWrapper(raw)


from __future__ import annotations

"""
Handler for Spawning Workflows
"""

import os
import time
import anyio
import signal
import atexit
import contextlib
import asyncio
import multiprocessing as mp
from lzl.logging import logger
from typing import Any, Dict, List, Optional, Type, Literal, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from .client import HatchetClient
    from .session import HatchetSession
    from .worker import Worker
    from .workflow import WorkflowT

_WorkflowTasks: Dict[str, Dict[str , asyncio.Task]] = {}
_WorkflowExitRegistered = False
_WorkflowLoop: Optional[asyncio.AbstractEventLoop] = None
_WorkflowEvent: Optional[asyncio.Event] = None

SIGNALS = [signal.SIGINT, signal.SIGTERM] if os.name != "nt" else [signal.SIGTERM]

def get_event_loop():
    """
    Returns the event loop
    """
    global _WorkflowLoop, _WorkflowEvent
    if not _WorkflowLoop:
        try:
            _WorkflowLoop = asyncio.get_running_loop()
        except RuntimeError as e:
            # logger.error(f"Error getting running loop: {e}")
            try:
                _WorkflowLoop = asyncio.get_event_loop()
            except RuntimeError as e:
                logger.error(f"Error getting event loop: {e}")
                _WorkflowLoop = asyncio.new_event_loop()
                asyncio.set_event_loop(_WorkflowLoop)
        
    if _WorkflowEvent is None:
        _WorkflowEvent = asyncio.Event()
        for signum in SIGNALS: _WorkflowLoop.add_signal_handler(signum, _WorkflowEvent.set)
    return _WorkflowLoop

# def create_event_loop():
#     """
#     Creates a new event loop
#     """
#     loop = asyncio.new_event_loop()
    
#     return loop


"""
V2 - Async Based Spawning
"""

def aexit_workflow_workers(timeout: Optional[int] = 5.0, workflow_names: Optional[List[str]] = None):
    """
    Exits all registered task workflow workers
    """
    global _WorkflowTasks
    workflow_names = workflow_names or list(_WorkflowTasks.keys())
    for workflow_name, wkflow_ops in _WorkflowTasks.items():
        if workflow_name not in workflow_names: continue
        for op_name, task in wkflow_ops.items():
            if task is None: continue
            task.cancel()
            try:
                task.result()
            except Exception as e:
                logger.error(f'[{workflow_name} - {op_name}] Error Stopping process: {e}')
                
            finally:
                task = None
    for workflow_name in workflow_names:
        if workflow_name in _WorkflowTasks:
            del _WorkflowTasks[workflow_name]


async def astart_worker_task(
    worker: 'Worker',
    workflow_name: str
):
    """
    Starts the worker task
    """
    worker.start()
    while not _WorkflowEvent.is_set():
        try:
            await asyncio.sleep(1)
        except Exception as e:
            logger.error(f'Exiting {workflow_name}: {e}')
            _WorkflowEvent.set()



async def astart_workflow(
    instance: Optional[str] = None,
    worker_name: Optional[str] = None,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    include_crontasks: Optional[bool] = None,
    only_crontasks: Optional[bool] = None,
    terminate_timeout: Optional[float] = 5.0,
    concurrency_limit: Optional[int] = None,
    extra_workflows: Optional[Dict[str, 'WorkflowT']] = None,
    **kwargs
    # method: str = 'v1',
):
    """
    Handles the start workflow
    """
    global _WorkflowExitRegistered, _WorkflowTasks
    from .client import HatchetClient
    hatchet = HatchetClient.get_session(instance = instance, **kwargs)
    if not worker_name: worker_name = f'{hatchet.instance}_workflows'
    if only_crontasks: worker_name += '.tasks'
    
    worker = hatchet.get_worker(worker_name, max_runs = concurrency_limit)
    workflows = hatchet.compile_workflows(include = include, exclude = exclude, extra_workflows = extra_workflows, include_crontasks = include_crontasks, only_crontasks = only_crontasks)
    logger.info(f'Compiled |g|{len(workflows)}|e| Workflows: {list(workflows.keys())}', colored = True, prefix = worker_name)
    for workflow in workflows.values():
        # logger.info(f'Starting Workflow: {workflow.op_name}: {workflow}', colored = True)
        worker.register_workflow(workflow)
    if not _WorkflowExitRegistered:
        atexit.register(aexit_workflow_workers, terminate_timeout)
        _WorkflowExitRegistered = True
    await worker.async_start()



async def run_workflow(
    instance: Optional[str] = None,
    worker_name: Optional[str] = None,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    include_crontasks: Optional[bool] = None,
    only_crontasks: Optional[bool] = None,
    seperate_workflows: Optional[bool] = None,
    terminate_timeout: Optional[float] = 5.0,
    concurrency_limit: Optional[int] = None,
    workflows: Optional[Dict[str, 'WorkflowT']] = None,

    workflow_name: Optional[str] = None,
    workflow_obj: Optional[Union[str, Type['WorkflowT']]] = None,
    workflow_mapping: Optional[Dict[str, Union[str, Type['WorkflowT']]]] = None,
    hatchet_client: Optional['HatchetClient'] = None,
    **kwargs
):
    # sourcery skip: reintroduce-else, remove-redundant-continue, replace-dict-items-with-values
    """
    Handles the start workflow
    """
    global _WorkflowExitRegistered, _WorkflowTasks
    if hatchet_client is None: 
        from .client import Hatchet
        hatchet_client = Hatchet
    # from .client import HatchetClient
    if workflow_mapping or worker_name or workflow_obj: hatchet_client.register_workflow(
        workflow_name = workflow_name,
        workflow_obj = workflow_obj,
        workflow_mapping = workflow_mapping,
        instance = instance,
        **kwargs,
    )

    hatchet = hatchet_client.get_session(instance = instance, **kwargs)
    if not worker_name: worker_name = f'{hatchet.instance}'
    loop = get_event_loop()
    # loop = asyncio.new_event_loop()
    workers: List['Worker'] = []
    worker = hatchet.get_worker(worker_name, max_runs = concurrency_limit)
    all_workflows: Dict[str, 'WorkflowT'] = {}
    if not only_crontasks:
        workflows = hatchet.compile_workflows(include = include, exclude = exclude, include_crontasks = False)
        logger.info(f'Compiled |g|{len(workflows)}|e| Workflows: {list(workflows.keys())}', colored = True, prefix = worker_name)
        for workflow_name, workflow in workflows.items():
            if workflow.is_disabled: continue
            worker.register_workflow(workflow)
            all_workflows[workflow_name] = workflow

        workers.append(worker)
    
    if include_crontasks or only_crontasks:
        if seperate_workflows:
            worker_name += '_tasks'
            task_worker = hatchet.get_worker(worker_name, max_runs = concurrency_limit)
        else:
            task_worker = worker
        task_workflows: Dict[str, 'WorkflowT'] = hatchet.compile_workflows(include = include, exclude = exclude, only_crontasks = True)
        logger.info(f'Compiled |g|{len(task_workflows)}|e| Cron Task Workflows: {list(task_workflows.keys())}', colored = True)
        for workflow_name, workflow in task_workflows.items():
            if workflow.is_disabled: continue
            if workflow.only_production_env and (
                not hatchet.hsettings.is_production_env
            ): 
                continue
            task_worker.register_workflow(workflow)
            all_workflows[workflow_name] = workflow
        if seperate_workflows:
            workers.append(task_worker)

    # loop = asyncio.new_event_loop()
    tasks: List[asyncio.Task] = [
        loop.create_task(worker.async_start()) for worker in workers
    ]

    async def exit_all_gracefully():
        exit_tasks = [loop.create_task(worker.exit_gracefully()) for worker in workers]
        # with anyio.move_on_after(10.0):
        await asyncio.gather(*exit_tasks)
        with contextlib.suppress(Exception):
            loop.stop()
    
    def exit_all_forcefully():
        for worker in workers:
            worker.exit_forcefully()

    loop.add_signal_handler(
        signal.SIGINT, lambda: asyncio.create_task(exit_all_gracefully())
    )
    loop.add_signal_handler(
        signal.SIGTERM, lambda: asyncio.create_task(exit_all_gracefully())
    )
    loop.add_signal_handler(signal.SIGQUIT, lambda: exit_all_forcefully())

    from hatchet_sdk.workflows_pb2 import CreateWorkflowVersionOpts
    for workflow_name, workflow in all_workflows.items():
        if workflow.workflow_overrides:
            hatchet.client.admin.put_workflow(
                f'{hatchet.hsettings.app_env.name}.{workflow_name}',
                workflow,
                overrides = CreateWorkflowVersionOpts(
                    **workflow.workflow_overrides
                ),
            )

    if not _WorkflowExitRegistered:
        atexit.register(aexit_workflow_workers, terminate_timeout)
        _WorkflowExitRegistered = True
    await asyncio.gather(*tasks)



def aexit_workflow_workers_v2(
    tasks: List['asyncio.Task'],
    timeout: Optional[int] = 5.0,     
):
    """
    Exits all registered task workflow workers
    """
    def inner(*args, **kwargs):
        logger.info('Exiting Workflows')
        for task in tasks:
            if not task.done():
                task.cancel()
            # try:
            #     task.result()
            # except Exception as e:
            #     logger.error(f'Error Stopping Workflow: {e}')
            # finally:
            #     task = None
    return inner

async def run_workers(
    tasks: List['asyncio.Task'],
):
    """
    Runs the workers
    """
    event = asyncio.Event()
    loop = asyncio.get_running_loop()
    for signum in SIGNALS: loop.add_signal_handler(signum, event.set)
    try:
        # await asyncio.wait(tasks)
        await event.wait()
    except Exception as e:
        logger.error(f'Error Stopping Workflow: {e}')
    finally:
        event.set()
        for task in tasks:
            task.cancel()
        await asyncio.wait_for(asyncio.gather(task), timeout = 5.0)


async def run_workflow_v2(
    instance: Optional[str] = None,
    worker_name: Optional[str] = None,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    include_crontasks: Optional[bool] = None,
    only_crontasks: Optional[bool] = None,
    seperate_workflows: Optional[bool] = None,
    terminate_timeout: Optional[float] = 5.0,
    concurrency_limit: Optional[int] = None,
    workflows: Optional[Dict[str, 'WorkflowT']] = None,

    workflow_name: Optional[str] = None,
    workflow_obj: Optional[Union[str, Type['WorkflowT']]] = None,
    workflow_mapping: Optional[Dict[str, Union[str, Type['WorkflowT']]]] = None,
    hatchet_client: Optional['HatchetClient'] = None,
    **kwargs
):
    # sourcery skip: reintroduce-else, remove-redundant-continue, replace-dict-items-with-values
    """
    Handles the start workflow
    """
    global _WorkflowExitRegistered, _WorkflowTasks
    if hatchet_client is None: 
        from .client import Hatchet
        hatchet_client = Hatchet
    # from .client import HatchetClient
    if workflow_mapping or worker_name or workflow_obj: hatchet_client.register_workflow(
        workflow_name = workflow_name,
        workflow_obj = workflow_obj,
        workflow_mapping = workflow_mapping,
        instance = instance,
        **kwargs,
    )

    hatchet = hatchet_client.get_session(instance = instance, **kwargs)
    if not worker_name: worker_name = f'{hatchet.instance}'
    workers: List['Worker'] = []
    worker = hatchet.get_worker(worker_name, max_runs = concurrency_limit)
    all_workflows: Dict[str, 'WorkflowT'] = {}
    if not only_crontasks:
        workflows = hatchet.compile_workflows(include = include, exclude = exclude, include_crontasks = False)
        logger.info(f'Compiled |g|{len(workflows)}|e| Workflows: {list(workflows.keys())}', colored = True, prefix = worker_name)
        for workflow_name, workflow in workflows.items():
            if workflow.is_disabled: continue
            worker.register_workflow(workflow)
            all_workflows[workflow_name] = workflow

        workers.append(worker)
    
    if include_crontasks or only_crontasks:
        if seperate_workflows:
            worker_name += '_tasks'
            task_worker = hatchet.get_worker(worker_name, max_runs = concurrency_limit)
        else:
            task_worker = worker
        task_workflows: Dict[str, 'WorkflowT'] = hatchet.compile_workflows(include = include, exclude = exclude, only_crontasks = True)
        logger.info(f'Compiled |g|{len(task_workflows)}|e| Cron Task Workflows: {list(task_workflows.keys())}', colored = True)
        for workflow_name, workflow in task_workflows.items():
            if workflow.is_disabled: continue
            if workflow.only_production_env and (
                not hatchet.hsettings.is_production_env
            ): 
                continue
            task_worker.register_workflow(workflow)
            all_workflows[workflow_name] = workflow
        if seperate_workflows:
            workers.append(task_worker)

    tasks: List[asyncio.Task] = [
        asyncio.create_task(worker.async_start()) for worker in workers
    ]
    from hatchet_sdk.workflows_pb2 import CreateWorkflowVersionOpts
    for workflow_name, workflow in all_workflows.items():
        if workflow.workflow_overrides:
            hatchet.client.admin.put_workflow(
                f'{hatchet.hsettings.app_env.name}.{workflow_name}',
                workflow,
                overrides = CreateWorkflowVersionOpts(
                    **workflow.workflow_overrides
                ),
            )

    atexit.register(aexit_workflow_workers_v2(tasks), terminate_timeout)
    # await run_workers(tasks)
    # await asyncio.gather(*tasks)


def build_workflows_v2(
    instance: Optional[str] = None,
    worker_name: Optional[str] = None,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    include_crontasks: Optional[bool] = None,
    only_crontasks: Optional[bool] = None,
    seperate_workflows: Optional[bool] = None,
    terminate_timeout: Optional[float] = 5.0,
    concurrency_limit: Optional[int] = None,
    workflows: Optional[Dict[str, 'WorkflowT']] = None,

    workflow_name: Optional[str] = None,
    workflow_obj: Optional[Union[str, Type['WorkflowT']]] = None,
    workflow_mapping: Optional[Dict[str, Union[str, Type['WorkflowT']]]] = None,
    hatchet_client: Optional['HatchetClient'] = None,
    **kwargs
) -> List['Worker']:
    # sourcery skip: reintroduce-else, remove-redundant-continue, replace-dict-items-with-values
    """
    Handles the start workflow
    """
    global _WorkflowExitRegistered, _WorkflowTasks
    if hatchet_client is None: 
        from .client import Hatchet
        hatchet_client = Hatchet
    # from .client import HatchetClient
    if workflow_mapping or worker_name or workflow_obj: hatchet_client.register_workflow(
        workflow_name = workflow_name,
        workflow_obj = workflow_obj,
        workflow_mapping = workflow_mapping,
        instance = instance,
        **kwargs,
    )

    hatchet = hatchet_client.get_session(instance = instance, **kwargs)
    if not worker_name: worker_name = f'{hatchet.instance}'
    workers: List['Worker'] = []
    worker = hatchet.get_worker(worker_name, max_runs = concurrency_limit)
    all_workflows: Dict[str, 'WorkflowT'] = {}
    if not only_crontasks:
        workflows = hatchet.compile_workflows(include = include, exclude = exclude, include_crontasks = False)
        logger.info(f'Compiled |g|{len(workflows)}|e| Workflows: {list(workflows.keys())}', colored = True, prefix = worker_name)
        for workflow_name, workflow in workflows.items():
            if workflow.is_disabled: continue
            worker.register_workflow(workflow)
            all_workflows[workflow_name] = workflow

        workers.append(worker)
    
    if include_crontasks or only_crontasks:
        if seperate_workflows:
            worker_name += '_tasks'
            task_worker = hatchet.get_worker(worker_name, max_runs = concurrency_limit)
        else:
            task_worker = worker
        task_workflows: Dict[str, 'WorkflowT'] = hatchet.compile_workflows(include = include, exclude = exclude, only_crontasks = True)
        logger.info(f'Compiled |g|{len(task_workflows)}|e| Cron Task Workflows: {list(task_workflows.keys())}', colored = True)
        for workflow_name, workflow in task_workflows.items():
            if workflow.is_disabled: continue
            if workflow.only_production_env and (
                not hatchet.hsettings.is_production_env
            ): 
                continue
            task_worker.register_workflow(workflow)
            all_workflows[workflow_name] = workflow
        if seperate_workflows:
            workers.append(task_worker)
    
    return workers
    tasks: List[asyncio.Task] = [
        asyncio.create_task(worker.async_start()) for worker in workers
    ]
    from hatchet_sdk.workflows_pb2 import CreateWorkflowVersionOpts
    for workflow_name, workflow in all_workflows.items():
        if workflow.workflow_overrides:
            hatchet.client.admin.put_workflow(
                f'{hatchet.hsettings.app_env.name}.{workflow_name}',
                workflow,
                overrides = CreateWorkflowVersionOpts(
                    **workflow.workflow_overrides
                ),
            )

    # atexit.register(aexit_workflow_workers_v2(tasks), terminate_timeout)


from __future__ import annotations

import functools
import dataclasses
import typing as t
import temporalio.converter
from temporalio.api.common.v1 import Payload
from lzl.proxied import ProxyObject
from temporalio.converter import (
    DataConverter,
    CompositePayloadConverter,
    DefaultPayloadConverter,
    EncodingPayloadConverter,
)
from lzl.types import eproperty
from lzl.io.ser.utils import is_primitive
from .utils import get_json_meta, get_serializer, get_json_loader, is_serializable


class SerializedEncodingPayloadConverter(EncodingPayloadConverter):
    
    def __init__(
        self, 
        jsonlib: t.Optional[str] = None,
        compression: t.Optional[str] = None, 
        compression_level: t.Optional[int] = None
    ):
        self._extra: t.Dict[str, t.Any] = {}
        self.jlib = get_serializer("json", enforce_byte_value = True, verbosity = -1, jsonlib = jsonlib, compression = compression, compression_level = compression_level)
        
    @eproperty
    def encoding(self) -> str:
        """
        Returns the encoding
        """
        # return "text/plain"
        return "json/object"
        # if self.jlib.compression_enabled: return f"{self.jlib.jsonlib_name}/compressed"
        # return f"{self.jlib.jsonlib_name}/plain"


    def to_payload(self, value: t.Any) -> t.Optional[Payload]:
        """
        Serializes the value
        """
        if not is_serializable(value): return None
        # print(f'{type(value)} to payload: ', value)
        # print('---' * 10)
        return Payload(
            metadata = get_json_meta(self.encoding, "serialized", self.jlib),
            data = self.jlib.dumps(
                value, 
                separators = (",", ":"), 
                sort_keys = True, 
            ),
        )

    def from_payload(self, payload: Payload, type_hint: t.Optional[t.Type] = None) -> t.Any:
        """
        Deserializes the value
        """
        if payload.metadata.get('converter') != b'serialized': return None
        # print('from payload: ', payload)
        # print('---' * 10)
        ser = get_json_loader(payload)
        return ser.loads(payload.data)


class SerializedPayloadConverter(CompositePayloadConverter):
    def __init__(self, jsonlib: t.Optional[str] = None, compression: t.Optional[str] = None, compression_level: t.Optional[int] = None, **kwargs) -> None:
        # Just add ours as first before the defaults
        super().__init__(
            SerializedEncodingPayloadConverter(jsonlib = jsonlib, compression = compression, compression_level = compression_level, **kwargs),
            *DefaultPayloadConverter.default_encoding_payload_converters,
        )

if t.TYPE_CHECKING:
    from lzl.ext.temporal.configs import TemporalSettings

def get_serialized_data_converter(
    jsonlib: t.Optional[str] = None,
    compression: t.Optional[str] = None,
    compression_level: t.Optional[int] = None,
    config: t.Optional['TemporalSettings'] = None,
    **kwargs
) -> DataConverter:
    """
    Creates a data converter that uses the serialized payload converter
    """
    if config is None: 
        from lzl.ext.temporal.configs import get_temporal_settings
        config = get_temporal_settings()
    if config.data_jsonlib is not None and jsonlib is None: jsonlib = config.data_jsonlib
    if config.data_compression is not None and compression is None: compression = config.data_compression
    if config.data_compression_level is not None and compression_level is None: compression_level = config.data_compression_level
    newobj = functools.partial(
        SerializedPayloadConverter,
        jsonlib = jsonlib,
        compression = compression,
        compression_level = compression_level,
        **kwargs
    )
    return dataclasses.replace(
        temporalio.converter.default(),
        payload_converter_class = newobj,
    )


serialized_data_converter: DataConverter = ProxyObject(
    obj_getter = get_serialized_data_converter,
)


from __future__ import annotations

from lzl.proxied import ProxyObject
from lzo.registry.settings import register_app_settings, get_app_settings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from lzl.io.file.configs.main import FileIOConfig

register_app_settings(
    'file',
    'lzl.io.file.configs.main.FileIOConfig',
)

def get_settings(**kwargs) -> 'FileIOConfig':
    """
    Returns the FileIO Settings
    """
    return get_app_settings('file', **kwargs)


fileio_settings: 'FileIOConfig' = ProxyObject(obj_getter = get_settings)


"""
This submodule only contains a compatiability layer for `pydantic` and `pydantic_settings`
"""

from .base import (
    BaseModel,
    validator,
    root_validator,
    pre_root_validator,
    Field,
    ConfigDict,
    model_validator,
    field_validator,
    get_schema_extra,
    ByteSize,
    PYDANTIC_VERSION,
    get_pydantic_field_names,
    PrivateAttr,
)
from .properties import (
    eproperty,
    aproperty,
    rproperty,
    lazyproperty,
    classproperty,
)

from .settings import (
    BaseSettings,
)

from .common import (
    Final, Literal, StrEnum, UpperStrEnum, AppEnv, get_app_env
)
from .typed import (
    AnyType,
    ClassType,
    SelfType,
    ParamType,
    ReturnType,
    LocalReturnType,
    CallableType,
    CallableAsyncType,
    CallableSyncOrAsyncType,
    CallableSyncOrAsyncReturnNoneType,
    MultiParamSpec,
    ProtocolParamType,
    ProtocolReturnType,
    ProtocolSelfType,
    # CallableAsyncNoParam,
    # CallableSyncNoParam,
    # CallableAsyncSingleParam,
    # CallableSyncSingleParam,
    # MethodAsyncNoParam,
    # MethodSyncNoParam,
    # MethodAsyncSingleParam,
    # MethodSyncSingleParam,
    AsyncGenerator
)

from .utils import create_alias_choices


from __future__ import annotations

from .base import (
    get_module_path,
    create_unique_secret_key,
    create_unique_id,
    update_dict,
    diff_dict,
    fail_after,
    is_coro_func,
    extract_function_kwargs,
)

from .batching import (
    split_into_batches,
    split_into_batches_with_index,
    split_into_batches_of_n,
    split_into_n_batches,
)

from .caching import (
    timed_cache,
    is_in_async_loop,
)

from .dates import (
    create_timestamp,
    is_expired_datetime,
    get_current_datetime,
    parse_datetime,
    parse_datetime_from_timestamp,
)

from .envvars import (
    parse_envvars_from_text,
    load_env_vars,
)

from .formatting import (
    build_dict_from_str,
    build_dict_from_list,
    build_dict_from_query,
    combine_parts,
)

from .timing import (
    Timer,
    timer,
    DynamicInterval,
)

from .extra import generate_json_schema_from_model


from __future__ import annotations

"""
Temporal Workflow Mixins: Workflows
"""

import typing as t
from .base import BaseTemporalMixin, MixinKinds

class TemporalWorkflowMixin(BaseTemporalMixin):
    """
    [Temporal] Workflow Mixin that will automatically registered
    """

    mixin_kind: t.Optional[MixinKinds] = 'workflow'
    _is_subclass_: t.Optional[bool] = True

    # These will be passed to the workflow defn decorator
    sandboxed: t.Optional[bool] = None
    dynamic: t.Optional[bool] = None
    failure_exception_types: t.Optional[t.Sequence[t.Type[BaseException]]] = None
    enable_init: t.Optional[bool] = None

    @classmethod
    def configure_registered(cls, **kwargs):
        """
        Configures the registered workflow
        """
        _kws = {k: kwargs.pop(k) for k in kwargs if k in {'sandboxed', 'dynamic', 'failure_exception_types'}}
        new = super().configure_registered(**kwargs)
        if 'sandboxed' in _kws: new.sandboxed = _kws.pop('sandboxed')
        if 'dynamic' in _kws: new.dynamic = _kws.pop('dynamic')
        if _kws.get('failure_exception_types'): new.failure_exception_types = _kws.pop('failure_exception_types')
        return new

    @classmethod
    def _on_register_hook_(cls):
        """
        Runs the workflow hook
        """
        pass


    def _on_init_hook_(self):
        """
        Runs the workflow init hook
        """
        pass



from __future__ import annotations

from starlette.authentication import BaseUser
from typing import Optional, Dict, Any, List, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from ..providers import ProviderClientT
    from ..manager import OAuth2Manager
    from .token import OAuth2Token
    

class OAuth2User(BaseUser):
    """
    The OAuth2 User
    """

    def __init__(
        self,
        token: Optional['OAuth2Token'],
        client: Optional['ProviderClientT'],
        manager: Optional['OAuth2Manager'],
        **kwargs,
    ):
        """
        Initializes the OAuth2 User
        """
        self.token = token
        self.client = client
        self.manager = manager
        self.is_admin = False
        self.is_staff = False
        self.roles = None

        
        # if self.token is not None:

    
    @property
    def is_authenticated(self) -> bool:
        """
        Returns True if the User is authenticated
        """
        return self.token is not None
    
    @property
    def display_name(self) -> Optional[str]:
        """
        Returns the User's Display Name
        """
        return self.token.display_name if self.is_authenticated else ""
    
    @property
    def email(self) -> Optional[str]:
        """
        Returns the User's Email
        """
        return self.token.email if self.is_authenticated else ""
    
    @property
    def identity(self) -> Optional[str]:
        """
        Returns the User's Identity
        """
        return self.token.identity if self.is_authenticated else ""
    
    async def alogout(self):
        """
        Logs the User Out
        """
        if not self.is_authenticated: return

    




from __future__ import annotations

"""
voyage.ai Reranking Route
"""

from ...rerankings import (
    RerankingRoute as BaseRerankingRoute,
    RerankingObject as BaseRerankingObject,
    RerankingResponse as BaseRerankingResponse,
    logger
)
from lzo.types import (
    validator, 
    model_validator,
    Field,
    eproperty
)
from lzl.api.openai.types.handlers import ModelContextHandler
from lzl.api.openai.types.base import Usage
from typing import Any, Dict, List, Optional, Union, Set, Type, TYPE_CHECKING

class RerankingObject(BaseRerankingObject):
    
    model: Optional[str] = "voyage-rerank-2-lite"
    return_documents: Optional[bool] = Field(False, description = 'If false, the API will return a list of {"index", "relevance_score"} where "index" refers to the index of a document within the input list., If true, the API will return a list of {"index", "document", "relevance_score"} where "document" is the corresponding document from the input list.') 
    truncation: Optional[bool] = Field(True, description = 'Whether to truncate the input text to the maximum context length')

class RerankingResponse(BaseRerankingResponse):

    usage: Optional[Usage] = Field(default_factory = Usage)

    @eproperty
    def consumption(self) -> int:
        """
        Returns the consumption for the reranking
        """
        try:
            return ModelContextHandler.get_consumption_cost(
                model_name = self.input_object.model,
                usage = self.usage,
            )
        except Exception as e:
            logger.error(f"Error getting consumption: {e}")
            return 0


class RerankingRoute(BaseRerankingRoute):
    input_model: Optional[Type[RerankingObject]] = RerankingObject
    response_model: Optional[Type[RerankingResponse]] = RerankingResponse

    @eproperty
    def create_enabled(self):
        """
        Returns whether the Create Route is Enabled
        """
        return True



from __future__ import annotations

from lzl.types.common import UpperStrEnum, _EXTEND_SUPPORTED
from typing import Union, Optional, Dict

if _EXTEND_SUPPORTED:
    from aenum import extend_enum

UserPrivilageLevel = {
    'ANON': 0,
    'UNKNOWN': 0,
    'AUTHENTICATED': 5,

    # Access Level (1-20)
    'READ': 5,
    'WRITE': 10,
    'MODIFY': 15,
    'PRIVILAGED': 20, # 'PRIVILAGED' is a user with 'READ', 'WRITE', and 'MODIFY' privilages
    
    # User Level (30-50)
    'USER': 30,             # 'USER' has 'READ', 'WRITE', and 'MODIFY' privilages
    'USER_API': 30,         # 'USER_API' is a user with 'USER' and 'API' privilages
    'USER_STAFF': 35,       # 'USER_STAFF' is a user with 'USER' and 'STAFF' privilages
    'USER_PRIVILAGED': 40,  # 'USER_PRIVILAGED' is a user with 'USER' and 'PRIVILAGED' privilages
    'USER_ADMIN': 45,       # 'USER_ADMIN' is a user with 'USER' and 'ADMIN' privilages
    
    # Internal Level (50-70)
    'STAFF': 50,            # 'STAFF' has 'READ', 'WRITE', and 'MODIFY' privilages
    'API_CLIENT': 50,       # 'API_CLIENT' is a user with 'API' privilages
    'SERVICE': 50,          # 'SERVICE' is a user with 'API' privilages
    'SERVICE_ACCOUNT': 50,  # 'SERVICE_ACCOUNT' is a user with 'API' privilages
    'SYSTEM': 60,           # 'SYSTEM' has 'READ', 'WRITE', and 'MODIFY' privilages
    
    # Admin Level (100+)
    'ADMIN': 100,           # 'ADMIN' has 'READ', 'WRITE', and 'MODIFY' privilages
    'SYSTEM_ADMIN': 150,
    'SUPER_ADMIN': 1000,
}
UserPrivilageIntLevel = {v: k for k, v in UserPrivilageLevel.items()}

DEFAULT_UNKNOWN_ROLE = 'UNKNOWN'
_DEFAULT_ROLE = object()

def set_default_unknown_role(role: Union[str, int, 'UserRole']) -> None:
    """
    Sets the default unknown role
    """
    global DEFAULT_UNKNOWN_ROLE
    if role != DEFAULT_UNKNOWN_ROLE:
        DEFAULT_UNKNOWN_ROLE = role
        from lazyops.utils.logs import logger
        logger.info(f'Set Default Unknown Role: {DEFAULT_UNKNOWN_ROLE}')

def extend_role(
    name: str,
    priviledge_level: int,
    set_as_default: bool = False,
    override: bool = False,
):
    """
    Extends the Role with a new level
    """
    if not _EXTEND_SUPPORTED: raise ImportError('aenum is not installed. Please install it to use this feature')
    global UserPrivilageLevel, UserPrivilageIntLevel
    name = name.upper()
    if name in UserPrivilageLevel:
        if not override: raise ValueError(f'Role {name} already exists')
    else:
        extend_enum(UserRole, name, name)
    
    UserPrivilageLevel[name] = priviledge_level
    UserPrivilageIntLevel[name] = priviledge_level
    if set_as_default: set_default_unknown_role(name)
    
def extend_roles(
    roles: Dict[str, int],
    set_as_default: Optional[str] = None,
    override: bool = False,
) -> None:
    """
    Extends the Roles with a new level
    """
    if not _EXTEND_SUPPORTED: raise ImportError('aenum is not installed. Please install it to use this feature')
    global UserPrivilageLevel, UserPrivilageIntLevel
    for role, priviledge_level in roles.items():
        if role in UserPrivilageLevel:
            if not override: raise ValueError(f'Role {role} already exists')
        else:
            extend_enum(UserRole, role, role)
        UserPrivilageLevel[role] = priviledge_level
        UserPrivilageIntLevel[role] = priviledge_level
        if set_as_default and role == set_as_default: set_default_unknown_role(role)

    
class UserRole(UpperStrEnum):
    ANON = 'ANON'
    UNKNOWN = 'UNKNOWN'
    # DEFAULT = 'DEFAULT'
    AUTHENTICATED = 'AUTHENTICATED'
    
    
    # Access Level (1-20)
    READ = 'READ'
    WRITE = 'WRITE'
    MODIFY = 'MODIFY'
    PRIVILAGED = 'PRIVILAGED'
    
    # User Level (30-50)
    USER = 'USER'
    USER_API = 'USER_API'
    USER_STAFF = 'USER_STAFF'
    USER_PRIVILAGED = 'USER_PRIVILAGED'
    USER_ADMIN = 'USER_ADMIN'

    # Internal Level (50-70)
    STAFF = 'STAFF'
    SYSTEM = 'SYSTEM'
    SERVICE = 'SERVICE'
    SERVICE_ACCOUNT = 'SERVICE_ACCOUNT'
    API_CLIENT = 'API_CLIENT'

    # Admin Level (100+)
    ADMIN = 'ADMIN'
    SYSTEM_ADMIN = 'SYSTEM_ADMIN'
    SUPER_ADMIN = 'SUPER_ADMIN'

    @property
    def privilage_level(self) -> int:
        """
        Returns the privilage level of the user role
        - Can be subclassed to return a different value
        """
        return UserPrivilageLevel[self.value]

    def __eq__(self, other: Union[int, str, 'UserRole']) -> bool:
        """
        Returns True if the user role is equal to the other
        """
        if other is None: other = UserRole.ANON
        if isinstance(other, int):
            return self.privilage_level == other
        if hasattr(other, 'privilage_level'):
            return self.privilage_level == other.privilage_level
        return self.privilage_level == UserPrivilageLevel[other.upper()]

    def __ne__(self, other: Union[int, str, 'UserRole']) -> bool:
        """
        Returns True if the user role is not equal to the other
        """
        if other is None: other = UserRole.ANON
        if isinstance(other, int):
            return self.privilage_level != other
        if hasattr(other, 'privilage_level'):
            return self.privilage_level != other.privilage_level
        return self.privilage_level != UserPrivilageLevel[other.upper()]

    def __lt__(self, other: Union[int, str, 'UserRole']) -> bool:
        """
        Returns True if the user role is less than the other
        """
        if other is None: other = UserRole.ANON
        if isinstance(other, int):
            return self.privilage_level < other
        if hasattr(other, 'privilage_level'):
            return self.privilage_level < other.privilage_level
        return self.privilage_level < UserPrivilageLevel[other.upper()]

    def __le__(self, other: Union[int, 'UserRole']) -> bool:
        """
        Returns True if the user role is less than or equal to the other
        """
        if other is None: other = UserRole.ANON
        if isinstance(other, int):
            return self.privilage_level <= other
        if hasattr(other, 'privilage_level'):
            return self.privilage_level <= other.privilage_level
        return self.privilage_level <= UserPrivilageLevel[other.upper()]

    def __gt__(self, other: Union[int, str, 'UserRole']) -> bool:
        """
        Returns True if the user role is greater than the other
        """
        if other is None: other = UserRole.ANON
        if isinstance(other, int):
            return self.privilage_level > other
        if hasattr(other, 'privilage_level'):
            return self.privilage_level > other.privilage_level
        return self.privilage_level > UserPrivilageLevel[other.upper()]
        
    def __ge__(self, other: Union[int, str, 'UserRole']) -> bool:
        """
        Returns True if the user role is greater than or equal to the other
        """
        if other is None: other = UserRole.ANON
        if isinstance(other, int):
            return self.privilage_level >= other
        if hasattr(other, 'privilage_level'):
            return self.privilage_level >= other.privilage_level
        return self.privilage_level >= UserPrivilageLevel[other.upper()]
    
    def __contains__(self, other: Union[int, str, 'UserRole']) -> bool:
        """
        Returns True if the user role is contained in the other
        """
        if other is None: other = UserRole.ANON
        if isinstance(other, int):
            return self.privilage_level >= other
        if hasattr(other, 'privilage_level'):
            return self.privilage_level >= other.privilage_level
        return self.privilage_level >= UserPrivilageLevel[other.upper()]

    @classmethod
    def parse_role(cls, role: Union[str, int], default: Optional[Union[str, int, 'UserRole']] = _DEFAULT_ROLE) -> 'UserRole':
        """
        Parses the role
        """
        if role is None: return UserRole.ANON
        try:
            return cls(UserPrivilageIntLevel[role]) if isinstance(role, int) else cls(role.upper())
        except Exception as e:
            from lazyops.utils.logs import logger
            if default is _DEFAULT_ROLE: default = DEFAULT_UNKNOWN_ROLE
            default_role = cls.parse_role(default, None) if default is not None else None
            logger.warning(f'Error Parsing Unknown Role: {role}: {e} - Returning {default_role} Role')
            return default_role
    
    def __hash__(self):
        """
        Returns the hash of the user role
        """
        return hash(self.value)


    @classmethod
    def set_default_unknown_role(cls, role: Union[str, int, 'UserRole']) -> None:
        """
        Sets the default unknown role
        """
        return set_default_unknown_role(role)

    @classmethod
    def extend_role(
        cls,
        name: str,
        priviledge_level: int,
        set_as_default: bool = False,
    ):
        """
        Extends the Role with a new level
        """
        return extend_role(name, priviledge_level, set_as_default)

    parse = parse_role


from __future__ import annotations

from starlette.middleware.authentication import AuthenticationBackend
from lzl.logging import logger
from typing import Optional, Dict, Any, List, Union, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from fastapi import Request, Response
    from starlette.requests import HTTPConnection
    from .config import OAuth2Settings
    from .types.user import OAuth2User
    from .types.token import OAuth2Token


class OAuth2Backend(AuthenticationBackend):
    """
    The OAuth2 Backend
    """
    def __init__(
        self,
        settings: 'OAuth2Settings',
        **kwargs,
    ):
        """
        The OAuth2 Backend
        """
        self.settings = settings
        self.manager = settings.get_manager()

    def get_jwt_token_from_request(self, request: 'Request') -> Optional[str]:
        """
        Gets the JWT Token from the Request
        """
        if not request.headers.get(self.settings.header): return None
        auth = request.headers.get(self.settings.header)
        scheme, token = auth.split(' ', 1)
        return None if scheme.lower() != self.settings.scheme.lower() else token

    async def has_valid_session(self, request: 'Request') -> bool:
        """
        Checks if the session is valid
        """
        if '_oauth' not in request.session: return False


    async def authenticate(
        self, 
        request: 'Request'
    ) -> Optional[Tuple["OAuth2Token", "OAuth2User"]]:
        """
        The authenticate method
        """
        # If a token is present, we always validate it first.
        # If no token is present, we check if there is a session.

        token = self.get_jwt_token_from_request(request)
        from_session = False
        if not token and '_oauth' in request.session:
            token = request.session['_oauth']
            from_session = True
            # logger.info(f'Authenticating Token from Session: {token}')
        if not token: return None, self.manager.user_class(None, None)
        return await self.manager.authenticate_oauth_token(token, request, from_session = from_session)
        
        # provider = await self.manager.get_provider_for_jwt(token, request)
        # logger.info(f'Authenticating Provider: {provider.name}')
        # token_payload = provider.parse_authorization_token(token, verify_exp=False)
        # # logger.info(f'Token Payload: {token_payload}')
        # oauth_token = self.manager.parse_token_payload(token_payload, provider)
        # if oauth_token.is_expired:
        # logger.info(oauth_token)
        # return oauth_token, self.manager.user_class(oauth_token, provider)






"""
Modular, Extensible Clients
"""

from .base import BaseClient
from .mixins import (
    CachifyClientMixin,
    CachifyHTTPClientMixin,
    BaseHTTPClientMixin,
    ResponseType,
    ResponseT,
    BrowserMixin,
)
from .http import BaseHTTPClient


from __future__ import annotations

"""
OpenAI Cost Functions and Handler
"""

import abc
import tiktoken
from lzl import load
from lzl.proxied import proxied, ProxyObject
from lzl.api.openai.assets import load_provider_prices
from .base import ModelCostItem, Usage

from typing import Optional, Union, Dict, Any, List, Tuple, Type, TYPE_CHECKING

if load.TYPE_CHECKING:
    import tiktoken
    import transformers
else:
    tiktoken = load.LazyLoad("tiktoken", install_missing=True)
    transformers = load.LazyLoad("transformers", install_missing=True)

if TYPE_CHECKING:
    from tiktoken import Encoding
    from lzl.api.openai.schemas.chat import ChatMessage
    from lzl.api.openai.configs.external import ExternalProviderSettings, ProviderModel
    # from async_openai.schemas.chat import ChatMessage
    from transformers import PreTrainedTokenizer
    # from async_openai.utils.external_config import ExternalProviderSettings, ProviderModel

# ModelCostHandlerClass
# @proxied

class ModelCostHandler(abc.ABC):
    """
    The Model Context Handler
    """
    def __init__(
        self,
        **kwargs,
    ):
        """
        Initializes the Model Cost Handler
        """
        self._models: Optional[Dict[str, ModelCostItem]] = None
        self._model_aliases: Optional[Dict[str, str]] = None
        self.tokenizers: Optional[Dict[str, 'Encoding']] = {}
        self.external_models: Optional[Dict[str, 'ProviderModel']] = {}
        self.external_model_aliases: Optional[Dict[str, str]] = {}
        self.external_tokenizers: Optional[Dict[str, 'PreTrainedTokenizer']] = {}
        self.default_for_routes: Optional[Dict[str, str]] = {}

    @staticmethod
    def load_models() -> Dict[str, ModelCostItem]:
        """
        Loads the models
        """
        model_prices = load_provider_prices(provider = 'openai')
        return {model['name']: ModelCostItem(**model) for model in model_prices}

    @property
    def models(self) -> Dict[str, ModelCostItem]:
        """
        Gets the models
        """
        if self._models is None: self._models = self.load_models()
        return self._models
    
    @property
    def model_aliases(self) -> Dict[str, str]:
        """
        Gets the model aliases
        """
        if self._model_aliases is None:
            self._model_aliases = {alias: model for model, item in self.models.items() for alias in item.aliases or []}
        return self._model_aliases


    def get_external_model(self, name: str) -> Optional['ProviderModel']:
        """
        Gets the model
        """
        if name not in self.external_model_aliases and name not in self.external_models:
            raise KeyError(f"Model {name} not found: {self.external_model_aliases} / {list(self.external_models.keys())}")
            # print(f"Model {name} not found: {self.external_model_aliases} / {self.external_models}")
            # return None
        if name in self.external_model_aliases:
            name = self.external_model_aliases[name]
        return self.external_models[name]
    
    def resolve_external_model_name(self, model_name: str) -> str:
        """
        Resolves the Model Name from the model aliases
        """
        # Try to remove the provider name
        model = self.get_external_model(model_name)
        return model.name

    def resolve_model_name(self, model_name: str) -> str:
        """
        Resolves the Model Name from the model aliases
        """
        # Try to remove the version number
        if model_name in self.models:
            return model_name
        
        # if model_name in self.model_aliases:
        return self.model_aliases.get(model_name, model_name)
        
    def __getitem__(self, key: str) -> ModelCostItem:
        """
        Gets a model by name
        """
        if '/' in key or key in self.external_model_aliases: return self.get_external_model(key)
        if key not in self.model_aliases and key not in self.models:
            return self.models[self.resolve_model_name(key)]
        if key in self.model_aliases:
            key = self.model_aliases[key]
        return self.models[key]
    
    def get(self, key: str, default: Optional[str] = None) -> Optional[ModelCostItem]:
        """
        Gets a model by name
        """
        try:
            return self[key]
        except KeyError:
            if default is None:
                raise KeyError(f"Model {key} not found") from None
            return self[default]
        
    def add_provider(self, provider: 'ExternalProviderSettings'):
        """
        Adds a provider to the handler
        """
        for model in provider.models:
            model_name = f'{provider.name}/{model.name}'
            self.external_models[model_name] = model
            if model.name not in self.external_model_aliases:
                self.external_model_aliases[model.name] = model_name
            for alias in model.aliases or []:
                model_alias = f'{provider.name}/{alias}'
                self.external_model_aliases[model_alias] = model_name
                if alias not in self.external_model_aliases:
                    self.external_model_aliases[alias] = model_name
            if not self.default_for_routes.get('rerankings') and model.supports('rerankings'):
                self.default_for_routes['rerankings'] = model_name
            elif model.data.get('is_default_for_rerankings', False):
                self.default_for_routes['rerankings'] = model_name
            
        
    def add_model(self, model: str, source_model: str):
        """
        Add a model to the handler
        
        Args:
            model (str): The model name
            source_model (str): The source model name
        """
        if model in self.model_aliases or model in self.models:
            return
        
        src_model = self[source_model]
        # Add to the model aliases
        self.model_aliases[model] = src_model.name

    def get_external_tokenizer(self, name: str) -> Optional['PreTrainedTokenizer']:
        """
        Gets the tokenizer
        """
        # Remove the provider name
        model = self.get_external_model(name)
        tokenizer_name = model.tokenizer or model.name
        if tokenizer_name not in self.external_tokenizers:
            # try:
            #     from transformers.models.auto.tokenization_auto import AutoTokenizer
            # except ImportError as e:
            #     raise ImportError("transformers is not installed, please install it to use this feature") from e            
            tokenizer = transformers.AutoTokenizer.from_pretrained(tokenizer_name)
            self.external_tokenizers[tokenizer_name] = tokenizer
        return self.external_tokenizers[tokenizer_name]

    def get_tokenizer(self, name: str) -> Optional[Union['Encoding', 'PreTrainedTokenizer']]:
        """
        Gets the tokenizer
        """
        # Switch the 35 -> 3.5
        # OpenAI Models don't have / in the name
        if '/' in name or name in self.external_model_aliases \
            or name in self.external_models: return self.get_external_tokenizer(name)
        
        if '35' in name: name = name.replace('35', '3.5')    
        if name not in self.tokenizers:
            if name in {'text-embedding-3-small', 'text-embedding-3-large'}:
                enc_name = 'cl100k_base'
                self.tokenizers[name] = tiktoken.get_encoding(enc_name)
            else:
                self.tokenizers[name] = tiktoken.encoding_for_model(name)
        return self.tokenizers[name]
    
    def count_chat_tokens(
        self, 
        messages: List[Union[Dict[str, str], 'ChatMessage']],
        model_name: str,
        reply_padding_token_count: Optional[int] = 3,
        message_padding_token_count: Optional[int] = 4,
        **kwargs
    ) -> int:
        """
        Returns the number of tokens in the chat.
        """
        num_tokens = 0
        tokenizer = self.get_tokenizer(model_name)
        for message in messages:
            if message.get('name'):
                num_tokens -= 1
            num_tokens += message_padding_token_count + len(tokenizer.encode(message.get('content', '')))
        num_tokens += reply_padding_token_count  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    def count_tokens(
        self,
        text: Union[str, List[str]],
        model_name: str,
        **kwargs
    ) -> int:
        """
        Returns the number of tokens in the text.
        """
        tokenizer = self.get_tokenizer(model_name)
        return (
            sum(len(tokenizer.encode(t)) for t in text)
            if isinstance(text, list)
            else len(tokenizer.encode(text))
        )
    
    def get_consumption_cost(self, model_name: str, usage: 'Usage', **kwargs) -> float:
        """
        Gets the consumption cost
        """
        # Switch the 35 -> 3.5
        if '35' in model_name: model_name = model_name.replace('35', '3.5')
        model = self[model_name]
        if isinstance(usage, dict):
            usage = Usage(**usage)
        return model.get_costs(usage = usage, **kwargs)
    
    
    def truncate_to_max_length(self, text: str, model_name: str, context_length: Optional[int] = None, **kwargs) -> str:
        """
        Truncates the text to the max length
        """
        tokenizer = self.get_tokenizer(model_name)
        if context_length is None:
            context_length = self[model_name].context_length
        tokens = tokenizer.encode(text)
        if len(tokens) > context_length:
            tokens = tokens[-context_length:]
            decoded = tokenizer.decode(tokens)
            text = text[-len(decoded):]
        return text

ModelContextHandler: ModelCostHandler = ProxyObject(ModelCostHandler)


import sys
import datetime
import inspect
import pathlib
import jinja2
import hatchet_sdk
from importlib.metadata import version

here = pathlib.Path(__file__).parent.parent
patches_dir = here.joinpath('_patches')

j2 = jinja2.Environment(loader=jinja2.FileSystemLoader(patches_dir))

last_version = '0.32.0'

replacements = {
    'worker.py': {
        '__init__': [
    # (
    #     'config: ClientConfig = {},',
    #     "config: ClientConfig = {},\n\
    #     session: Optional['HatchetSession'] = None,\n\
    #     context_class: Type[Context] = Context,\n"
    # ),
    (
        'labels: dict[str, str | int] = {},',
        "labels: dict[str, str | int] = {},\n\
        session: Optional['HatchetSession'] = None,\n\
        context_class: Type[Context] = Context,\n"
    ),
    (
        'self.client = new_client_raw(config)', 
        "self.session = session\n\
        self.client = self.session.client if self.session else new_session(config)\n\
        self._loop: Optional[asyncio.AbstractEventLoop] = None\n\
        self._context_cls = context_class\n\
        self.main_task = None\n"
    )
        ],
        'handle_start_step_run': [
    (
        'context = Context(',
        'context = self._context_cls(',
    )
        ],
        'handle_start_group_key_run': [
    (
        'context = Context(',
        'context = self._context_cls(',
    ),
    (
        'except Exception as e:',
            'except Exception as e:\n\
                if "]:" in str(e) and "- 40" in str(e):\n\
                    logger.error(f"[{action.step_run_id} - {action.job_name}] Error in action: {e}")\n\
                    raise e',
        1
    )
        ],
        'async_wrapped_action_func': [
    (
        'except Exception as e:',
        'except Exception as e:\n\
            if "]:" in str(e) and "- 40" in str(e):\n\
                logger.error(f"[{action.step_run_id} - {action.job_name}] Error in action: {e}")\n\
                raise e'
    )
        ],
        'get_step_action_finished_event': [
    (
        'json.loads',
        'json_serializer.loads'
    ),
        ],
    #     '_async_start': [
    # (
    #     'self.loop = asyncio.get_running_loop()',
    #     'self.loop = asyncio.get_running_loop()\n\
    #     self.run_watcher_event()',
    #     # '# self.loop = asyncio.get_running_loop()',
    # )
    #     ]
    },
    'context.py': {
        '__init__': [
            (
                'json.loads',
                'json_serializer.loads'
            ),
            (
                'Action',
                "'Action'", 1
            ),
            (
                'DispatcherClientImpl',
                "'DispatcherClientImpl'", 1
            ),
            (
                'AdminClientImpl',
                "'AdminClientImpl'", 1
            ),
            (
                'EventClientImpl',
                "'EventClientImpl'", 1
            ),
            (
                'PooledWorkflowRunListener',
                "'PooledWorkflowRunListener'", 1
            ),
            (
                'RunEventListenerClient',
                "'RunEventListenerClient'", 1
            ),
            (
                'WorkerContext',
                "'WorkerContext'", 1
            )
        ],
    },

}


def patch_worker(
    version: str,
    timestamp: str,
    version_dir: pathlib.Path,
    version_id: float,
):
    """
    Patches the worker class to add the ability to handle start step run
    """
    worker = hatchet_sdk.worker.Worker
    patched_funcs = {}
    for name, patches in replacements['worker.py'].items():
        func_string = inspect.getsource(getattr(worker, name))
        for patch in patches:
            func_string = func_string.replace(*patch)
        patched_funcs[name] = func_string
    
    template = j2.get_template('worker.j2.py')
    output = template.render(new_funcs = patched_funcs, version = version, timestamp = timestamp, last_version = last_version)
    output_file = version_dir.joinpath('worker.py')
    output_file.write_text(output)

def patch_context(
    version: str,
    timestamp: str,
    version_dir: pathlib.Path,
    version_id: float,
):
    """
    Patches the context class to add the necessary methods
    """
    # context = hatchet_sdk.context.context.Context
    from hatchet_sdk.context.context import Context
    context = Context
    patched_funcs = {}
    for name, patches in replacements['context.py'].items():
        func_string = inspect.getsource(getattr(context, name))
        for patch in patches:
            func_string = func_string.replace(*patch)
        patched_funcs[name] = func_string
    
    template = j2.get_template('context.j2.py')
    output = template.render(new_funcs = patched_funcs, version = version, timestamp = timestamp, last_version = last_version)
    output_file = version_dir.joinpath('context.py')
    output_file.write_text(output)

def create_version_file(
    version: str,
    version_dir: pathlib.Path,
    version_id: float,
):
    """
    Creates the version file
    """
    template = j2.get_template('version.j2.py')
    output = template.render(version = version, last_version = last_version)
    output_file = version_dir.joinpath('version.py')
    output_file.write_text(output)

def create_init_file(
    version: str,
    version_dir: pathlib.Path,
    version_id: float,
) -> None:
    """
    Creates the init file
    """
    template = j2.get_template('ver_init.j2.py')
    output = template.render(version = version, last_version = last_version)
    output_file = version_dir.joinpath('__init__.py')
    output_file.write_text(output)

def create_current_file(
    version: str,
    version_dir: pathlib.Path,
    version_id: float,
) -> None:
    """
    Creates the current file
    """
    template = j2.get_template('current.j2.py')
    current_version_name = version_dir.name
    output = template.render(version = version, last_version = last_version, current_version_name = current_version_name)
    output_file = here.joinpath('current', '__init__.py')
    output_file.write_text(output)


def run_patches():
    ver = version('hatchet-sdk')
    timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    ver_dir_name = f'v{ver.replace(".", "")}'
    ver_dir = here.joinpath(ver_dir_name)
    ver_dir.mkdir(exist_ok = True)
    ver_id = float(ver.split('.', 1)[1])
    patch_worker(ver, timestamp, ver_dir, ver_id)
    patch_context(ver, timestamp, ver_dir, ver_id)
    create_version_file(ver, ver_dir, ver_id)
    create_init_file(ver, ver_dir, ver_id)
    create_current_file(ver, ver_dir, ver_id)

if __name__ == '__main__':
    run_patches()

# patch_worker()
# patch_context()



from __future__ import annotations

import os
import inspect
import typing as t
# import aiopath.selectors
# import aiopath.scandir
from aiopath.wrap import func_to_async_func
from hashlib import md5
try:
    import xxhash
    _xxhash_available = True
except ImportError:
    xxhash = None
    _xxhash_available = False


if t.TYPE_CHECKING:
    from ..compat._aiopath.scandir import EntryWrapper
    # from aiopath.scandir import EntryWrapper
    from .base import FilePath


def scandir_sync(*args, **kwargs) -> t.Iterable['EntryWrapper']:
    """
    Sync version of scandir
    """
    from ..compat._aiopath.scandir import _scandir_results
    yield from _scandir_results(*args, **kwargs)


close = func_to_async_func(os.close)
sync_close = os.close


def generate_checksum_xxhash(p: 'FilePath', chunk_size: int = 8192) -> str:
    """
    Generates a checksum
    """
    if not _xxhash_available:
        raise RuntimeError("xxhash is not available")
    with p.open('rb') as f:
        file_hash = xxhash.xxh3_128()
        chunk = f.read(chunk_size)
        while chunk:
            file_hash.update(chunk)
            chunk = f.read(chunk_size)
    return file_hash.hexdigest()



def generate_checksum(p: 'FilePath', chunk_size: int = 8192) -> str:
    """
    Generates a checksum
    """
    if _xxhash_available: return generate_checksum_xxhash(p, chunk_size)
    with p.open('rb') as f:
        file_hash = md5()
        chunk = f.read(8192)
        while chunk:
            file_hash.update(chunk)
            chunk = f.read(8192)
    return file_hash.hexdigest()


def calc_etag(inputfile: 'FilePath', partsize: int = 8388608):
    """
    Calculate the etag
    """
    md5s = []
    with open(inputfile,'rb') as fp:
        while True:
            data = fp.read(partsize)
            if not data: break
            md5s.append(md5(data))
    if len(md5s) > 1:
        digests = b"".join(m.digest() for m in md5s)
        new_md5 = md5(digests)
        return f'"{new_md5.hexdigest()}-{len(md5s)}"'
    elif len(md5s) == 1: # file smaller than chunk size
        return f'"{md5s[0].hexdigest()}"'
    else: # empty file
        return '""'


    # md5_digests = []
    # with inputfile.open('rb') as f:
    #     md5_digests.extend(md5(chunk).digest() for chunk in iter(lambda: f.read(partsize), b''))
    # return f"{md5(b''.join(md5_digests)).hexdigest()}-{len(md5_digests)}"


def iscoroutinefunction(obj):
    if inspect.iscoroutinefunction(obj): return True
    return bool(hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__))



from __future__ import annotations

"""
Base Jinja2 Extension
"""


import jinja2
import typing as t
from jinja2 import FileSystemLoader
from lzl.types import eproperty
from lzl.types import Literal
from .filters import autofilters
from typing import Optional, Generic, TypeVar, TYPE_CHECKING, overload, Dict, Any

if TYPE_CHECKING:
    from jinja2.environment import Template
    from jinja2.environment import (
        BLOCK_START_STRING,
        BLOCK_END_STRING,
        VARIABLE_START_STRING,
        VARIABLE_END_STRING,
        COMMENT_START_STRING,
        COMMENT_END_STRING,
        LINE_STATEMENT_PREFIX,
        LINE_COMMENT_PREFIX,
        TRIM_BLOCKS,
        LSTRIP_BLOCKS,
        NEWLINE_SEQUENCE,
        KEEP_TRAILING_NEWLINE,
        Extension,
        Undefined,
        BaseLoader,
        BytecodeCache,
    )

J2EnvT = TypeVar('J2EnvT', bound = jinja2.Environment)

class Environment(Generic[J2EnvT]):

    include_autofilters: Optional[bool] = True
    default_mode: Optional[Literal['async', 'sync']] = 'sync'


    _extra: Dict[str, Any] = {}

    
    def __init__(
        self,
        include_autofilters: Optional[bool] = None,
        default_mode: Optional[Literal['async', 'sync']] = None,
        **kwargs
    ):
        """
        Initializes the Environment
        """
        if include_autofilters is not None: self.include_autofilters = include_autofilters
        if default_mode is not None: self.default_mode = default_mode
        _ = kwargs.pop('enable_async', None)
        self._env_kwargs = kwargs
    

    if TYPE_CHECKING:

        def __init__(
            self,
            block_start_string: str = BLOCK_START_STRING,
            block_end_string: str = BLOCK_END_STRING,
            variable_start_string: str = VARIABLE_START_STRING,
            variable_end_string: str = VARIABLE_END_STRING,
            comment_start_string: str = COMMENT_START_STRING,
            comment_end_string: str = COMMENT_END_STRING,
            line_statement_prefix: t.Optional[str] = LINE_STATEMENT_PREFIX,
            line_comment_prefix: t.Optional[str] = LINE_COMMENT_PREFIX,
            trim_blocks: bool = TRIM_BLOCKS,
            lstrip_blocks: bool = LSTRIP_BLOCKS,
            newline_sequence: Literal['\\n', '\\r\\n', '\\r'] = NEWLINE_SEQUENCE, 
            keep_trailing_newline: bool = KEEP_TRAILING_NEWLINE,
            extensions: t.Sequence[t.Union[str, t.Type["Extension"]]] = (),
            optimized: bool = True,
            undefined: t.Type[Undefined] = Undefined,
            finalize: t.Optional[t.Callable[..., t.Any]] = None,
            autoescape: t.Union[bool, t.Callable[[t.Optional[str]], bool]] = False,
            loader: t.Optional["BaseLoader"] = None,
            cache_size: int = 400,
            auto_reload: bool = True,
            bytecode_cache: t.Optional["BytecodeCache"] = None,
            include_autofilters: Optional[bool] = None,
            default_mode: Optional[Literal['async', 'sync']] = None,
        ):
            r"""The core component of Jinja is the `Environment`.  It contains
            important shared variables like configuration, filters, tests,
            globals and others.  Instances of this class may be modified if
            they are not shared and if no template was loaded so far.
            Modifications on environments after the first template was loaded
            will lead to surprising effects and undefined behavior.

            Here are the possible initialization parameters:

                `block_start_string`
                    The string marking the beginning of a block.  Defaults to ``'{%'``.

                `block_end_string`
                    The string marking the end of a block.  Defaults to ``'%}'``.

                `variable_start_string`
                    The string marking the beginning of a print statement.
                    Defaults to ``'{{'``.

                `variable_end_string`
                    The string marking the end of a print statement.  Defaults to
                    ``'}}'``.

                `comment_start_string`
                    The string marking the beginning of a comment.  Defaults to ``'{#'``.

                `comment_end_string`
                    The string marking the end of a comment.  Defaults to ``'#}'``.

                `line_statement_prefix`
                    If given and a string, this will be used as prefix for line based
                    statements.  See also :ref:`line-statements`.

                `line_comment_prefix`
                    If given and a string, this will be used as prefix for line based
                    comments.  See also :ref:`line-statements`.

                    .. versionadded:: 2.2

                `trim_blocks`
                    If this is set to ``True`` the first newline after a block is
                    removed (block, not variable tag!).  Defaults to `False`.

                `lstrip_blocks`
                    If this is set to ``True`` leading spaces and tabs are stripped
                    from the start of a line to a block.  Defaults to `False`.

                `newline_sequence`
                    The sequence that starts a newline.  Must be one of ``'\r'``,
                    ``'\n'`` or ``'\r\n'``.  The default is ``'\n'`` which is a
                    useful default for Linux and OS X systems as well as web
                    applications.

                `keep_trailing_newline`
                    Preserve the trailing newline when rendering templates.
                    The default is ``False``, which causes a single newline,
                    if present, to be stripped from the end of the template.

                    .. versionadded:: 2.7

                `extensions`
                    List of Jinja extensions to use.  This can either be import paths
                    as strings or extension classes.  For more information have a
                    look at :ref:`the extensions documentation <jinja-extensions>`.

                `optimized`
                    should the optimizer be enabled?  Default is ``True``.

                `undefined`
                    :class:`Undefined` or a subclass of it that is used to represent
                    undefined values in the template.

                `finalize`
                    A callable that can be used to process the result of a variable
                    expression before it is output.  For example one can convert
                    ``None`` implicitly into an empty string here.

                `autoescape`
                    If set to ``True`` the XML/HTML autoescaping feature is enabled by
                    default.  For more details about autoescaping see
                    :class:`~markupsafe.Markup`.  As of Jinja 2.4 this can also
                    be a callable that is passed the template name and has to
                    return ``True`` or ``False`` depending on autoescape should be
                    enabled by default.

                    .. versionchanged:: 2.4
                    `autoescape` can now be a function

                `loader`
                    The template loader for this environment.

                `cache_size`
                    The size of the cache.  Per default this is ``400`` which means
                    that if more than 400 templates are loaded the loader will clean
                    out the least recently used template.  If the cache size is set to
                    ``0`` templates are recompiled all the time, if the cache size is
                    ``-1`` the cache will not be cleaned.

                    .. versionchanged:: 2.8
                    The cache size was increased to 400 from a low 50.

                `auto_reload`
                    Some loaders load templates from locations where the template
                    sources may change (ie: file system or database).  If
                    ``auto_reload`` is set to ``True`` (default) every time a template is
                    requested the loader checks if the source changed and if yes, it
                    will reload the template.  For higher performance it's possible to
                    disable that.

                `bytecode_cache`
                    If set to a bytecode cache object, this object will provide a
                    cache for the internal Jinja bytecode so that templates don't
                    have to be parsed if they were not changed.

                    See :ref:`bytecode-cache` for more information.
            """



    @eproperty
    def env(self) -> J2EnvT:
        """
        [Non-Async] Returns the Environment
        """
        new_env = jinja2.Environment(**self._env_kwargs)
        if self.include_autofilters:
            new_env.filters.update(autofilters)
        return new_env
    
    @eproperty
    def aenv(self) -> J2EnvT:
        """
        [Async] Returns the Environment
        """
        new_env = jinja2.Environment(**self._env_kwargs, enable_async = True)
        if self.include_autofilters:
            new_env.filters.update(autofilters)
        return new_env
    
    def _get_env(self, mode: Optional[Literal['async', 'sync']] = None) -> J2EnvT:
        """
        Returns the Environment
        """
        if mode is None: mode = self.default_mode
        return self.aenv if mode == 'async' else self.env
    
    def _run_env_func(self, func: str, *args, **kwargs) -> Any:
        """
        Runs a function on the Environment
        """
        getattr(self.env, func)(*args, **kwargs)
        getattr(self.aenv, func)(*args, **kwargs)

            
    def get_template(
        self,
        name: t.Union[str, "Template"],
        parent: t.Optional[str] = None,
        globals: t.Optional[t.MutableMapping[str, t.Any]] = None,
        mode: t.Optional[Literal['async', 'sync']] = None,
        is_raw: t.Optional[bool] = None,
    ) -> "Template":
        """Load a template by name with :attr:`loader` and return a
        :class:`Template`. If the template does not exist a
        :exc:`TemplateNotFound` exception is raised.

        :param name: Name of the template to load. When loading
            templates from the filesystem, "/" is used as the path
            separator, even on Windows.
        :param parent: The name of the parent template importing this
            template. :meth:`join_path` can be used to implement name
            transformations with this.
        :param globals: Extend the environment :attr:`globals` with
            these extra variables available for all renders of this
            template. If the template has already been loaded and
            cached, its globals are updated with any new items.

        .. versionchanged:: 3.0
            If a template is loaded from cache, ``globals`` will update
            the template's globals instead of ignoring the new values.

        .. versionchanged:: 2.4
            If ``name`` is a :class:`Template` object it is returned
            unchanged.
        """
        _env = self._get_env(mode)
        if is_raw:  return _env.from_string(name)
        return _env.get_template(name, parent, globals)

    def add_extension(self, extension: t.Union[str, t.Type["Extension"]]) -> None:
        """Adds an extension after the environment was created.

        .. versionadded:: 2.5
        """
        self._run_env_func('add_extension', extension)

    def extend(self, **attributes: t.Any) -> None:
        """Add the items to the instance of the environment if they do not exist
        yet.  This is used by :ref:`extensions <writing-extensions>` to register
        callbacks and configuration values without breaking inheritance.
        """
        self._run_env_func('extend', **attributes)

    


from __future__ import annotations

import typing as t
from lzl.api import aiohttpx
from lzl.api.searxng.models import ResultItem


"""
https://github.com/BrainBlend-AI/atomic-agents/blob/main/atomic-examples/orchestration-agent/orchestration_agent/tools/searxng_search.py
"""


from __future__ import annotations

"""
The Hatchet Version Init

Patched: {{ timestamp }}
Version: {{ version }}
Last Version: {{ last_version }}
"""

from .context import Context, ContextT
from .worker import Worker
from .version import CURRENT_VERSION, LAST_VERSION


from __future__ import annotations

"""
Batching Helpers
"""

import itertools
import typing
from typing import Callable, Dict, Any, Tuple, TypeVar, Generic, Iterable, List, Generator, TYPE_CHECKING

ItemT = TypeVar('ItemT')

def get_batches_from_generator(iterable: typing.Iterable[ItemT], n: int) -> typing.Generator[List[ItemT], None, None]:
    """
    Batch elements of an iterable into fixed-length chunks or blocks.
    """
    it = iter(iterable)
    while x := tuple(itertools.islice(it, n)):
        yield x


def split_into_batches(items: List[ItemT], n: int) -> typing.Iterable[typing.List[ItemT]]:
    """
    Splits the items into n amount of equal items

    >>> list(split_into_batches(range(11), 3))
    [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10]]
    """
    n = min(n, len(items))
    k, m = divmod(len(items), n)
    return (items[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(n))

def split_into_batches_of_n(iterable: typing.Iterable[ItemT], n: int) -> typing.Iterable[typing.List[ItemT]]:
    """
    Splits the items into fixed-length chunks or blocks.

    >>> list(split_into_batches_of_n(range(11), 3))
    [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10]]
    """
    return list(get_batches_from_generator(iterable, n))


def split_into_n_batches(iterable: typing.Iterable[ItemT], size: int) -> typing.Iterable[typing.List[ItemT]]:
    """
    Splits the items into n amount of equal items

    >>> list(split_into_batches_of_size(range(11), 3))
    [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10]]
    """
    return split_into_batches(iterable, size)


def build_batches(iterable: typing.Iterable[ItemT], size: int, fixed_batch_size: bool = True) -> typing.Iterable[typing.List[ItemT]]:
    """
    Builds batches of a given size from an iterable.
    """
    if fixed_batch_size:
        return split_into_batches_of_n(iterable, size)
    return split_into_n_batches(iterable, size)


def split_into_batches_with_index(iterable: typing.Iterable[ItemT], n: int, start: typing.Optional[int] = None) -> typing.Iterable[typing.Tuple[int, List[ItemT]]]:
    """
    Splits the items into fixed-length chunks or blocks.

    >>> list(split_into_batches_of_n(range(11), 3))
    [(0, [0, 1, 2]), (1, [3, 4, 5]), (2, [6, 7, 8]), (3, [9, 10])]
    """
    return list(enumerate(get_batches_from_generator(iterable, n), start = start or 0))



from __future__ import annotations

"""
voyage.ai Embedding Route
"""


from ...embeddings import (
    EmbeddingRoute as BaseEmbeddingRoute,
    EmbeddingObject as BaseEmbeddingObject,
    EmbeddingResponse as BaseEmbeddingResponse,
    logger
)
from lzo.types import (
    validator, 
    model_validator,
    Field,
    eproperty
)
from lzl.api.openai.types.handlers import ModelContextHandler
from lzl.api.openai.types.base import Usage
from typing import Any, Dict, List, Optional, Union, Set, Type, TYPE_CHECKING

_allowed_dtypes = [
    'float', 'int8', 'uint8', 'binary', 'ubinary'
]

class EmbeddingObject(BaseEmbeddingObject):
    
    model: Optional[str] = "voyage-3-lite"
    truncation: Optional[bool] = Field(True, description = "Whether to truncate the input text to the maximum context length")
    output_dimension: Optional[int] = Field(None, description = "The output dimension of the embedding")
    output_dtype: Optional[str] = Field(None, description = "The output data type of the embedding")

    # Discarded Properties
    user: Optional[str] = Field(None, exclude = True, hidden = True)
    dimensions: Optional[int] = Field(None, exclude = True, hidden = True)

    @validator('model', pre=True, always=True)
    def validate_model(cls, v, values: Dict[str, Any]) -> str:
        """
        Validate the model
        """
        if not v:
            if values.get('engine'):
                v = values.get('engine')
            elif values.get('deployment'):
                v = values.get('deployment')
        v = ModelContextHandler.resolve_external_model_name(v)
        return v
    
    @model_validator(mode = 'after')
    def validate_embedding_object(self):
        """
        Validate the embedding object
        """
        if self.output_dimension is not None:
            c = ModelContextHandler.get_external_model(self.model)
            if c.data.get('dims') and self.output_dimension not in c.data['dims']:
                raise ValueError(f"Invalid Output Dimension: {self.output_dimension}. Must be one of {c.data['dims']}")
        if self.output_dtype is not None:
            if self.output_dtype not in _allowed_dtypes:
                raise ValueError(f"Invalid Output Data Type: {self.output_dtype}. Must be one of {_allowed_dtypes}")
        return self


class EmbeddingResponse(BaseEmbeddingResponse):

    usage: Optional[Usage] = Field(default_factory = Usage)


    @eproperty
    def consumption(self) -> int:
        """
        Returns the consumption for the completions
        """
        try:
            if not self.usage.prompt_tokens:
                self.usage.prompt_tokens = ModelContextHandler.count_tokens(self.input_object.input, model_name=self.input_object.model)
            return ModelContextHandler.get_consumption_cost(
                model_name = self.input_object.model,
                usage = self.usage,
            )
        except Exception as e:
            logger.error(f"Error getting consumption: {e}")
            return 0


class EmbeddingRoute(BaseEmbeddingRoute):
    input_model: Optional[Type[EmbeddingObject]] = EmbeddingObject
    response_model: Optional[Type[EmbeddingResponse]] = EmbeddingResponse


from __future__ import annotations

import fnmatch
import datetime
import typing as t
import json
from fastapi import HTTPException
from lzo.types import BaseModel, eproperty, Field
from lzl.logging import logger
from pydantic import PrivateAttr, model_validator, field_validator, AliasChoices, field_serializer, ConfigDict, AliasGenerator
from pydantic.alias_generators import to_camel, to_snake
from pydantic import ValidationInfo

def validation_gen(field: str) -> AliasChoices:
    """
    Generate validation aliases for a field.
    """
    return AliasChoices(to_camel(field), field)

default_alias_gen = AliasGenerator(
    validation_alias=validation_gen,
    serialization_alias=to_camel,
)

_DefaultKeycloakClientIds = [
    'account',
    'account-console',
    'admin-cli',
    'admin-portal',
    'broker',
    'idp-wizard',
    'realm-management',
    'security-admin-console'
]

_ValidEventTypes = {
    "LOGIN",
    "LOGOUT",
    "USER_INFO_REQUEST",
}

class BaseKCModel(BaseModel):
    """
    Base Keycloak Model
    """
    # The other stuff we don't really care about
    model_config = ConfigDict(
        extra = "allow",
        arbitrary_types_allowed=True,
        alias_generator=default_alias_gen,
        serialize_by_alias=True,
    )


class KeycloakRealm(BaseKCModel):
    """
    Represents a Keycloak realm.
    """
    id: str
    realm: str
    display_name: str = Field(...)

    @eproperty
    def name(self) -> str:
        """
        Returns the realm name
        """
        return self.realm


class ProtocolMapperItem(BaseKCModel):
    id: str
    name: t.Optional[str] = Field(None)
    protocol: t.Optional[str] = Field(None)
    protocol_mapper: t.Optional[str] = Field(None)
    consent_required: t.Optional[bool] = Field(None)
    config: t.Dict[str, t.Any] = Field(default_factory=dict)


class KeycloakClient(BaseKCModel):
    """
    Represents a Keycloak client.
    """
    id: str
    client_id: str = Field(...)
    name: str = Field(...)
    root_url: t.Optional[str] = Field(None)
    base_url: t.Optional[str] = Field(None)

    admin_url: t.Optional[str] = Field(None)
    enabled: t.Optional[bool] = Field(default=True)
    
    client_authenticator_type: t.Optional[str] = Field(None)
    secret: t.Optional[str] = Field(None)

    web_origins: t.List[str] = Field(default_factory=list)
    redirect_uris: t.List[str] = Field(default_factory=list)

    protocol: t.Optional[str] = Field(None)
    attributes: t.Optional[t.Dict[str, t.Any]] = Field(default_factory=dict)
    protocol_mappers: t.List[ProtocolMapperItem] = Field(default_factory=list)
    default_client_scopes: t.List[str] = Field(default_factory=list)
    optional_client_scopes: t.List[str] = Field(default_factory=list)
    access: t.Optional[t.Dict[str, bool]] = Field(default_factory=dict)


    @eproperty
    def is_default_client_id(self) -> bool:
        """
        Checks if the client ID is a default Keycloak client ID.
        """
        return self.client_id in _DefaultKeycloakClientIds


class KeycloakUser(BaseKCModel):
    """
    Represents a Keycloak user.
    """
    id: str
    username: str
    first_name: t.Optional[str] = Field(None)
    last_name: t.Optional[str] = Field(None)
    email: t.Optional[str] = None
    email_verified: bool = Field(default=False)
    enabled: bool = Field(default=True)
    created_timestamp: t.Optional[int] = Field(None)
    totp: bool = Field(default=False)
    disableable_credential_types: t.List[str] = Field(default_factory=list)
    required_actions: t.List[str] = Field(default_factory=list)
    not_before: int = Field(default=0)
    access: t.Optional[t.Dict[str, t.Any]] = None
    attributes: t.Optional[t.Dict[str, t.List[str] | t.Dict[str, str]]] = Field(default_factory=dict, validation_alias=AliasChoices("attributes", "userAttributes", "user_attributes"))

    @field_validator("attributes", mode='before')
    def validate_attributes(cls, v: t.Optional[t.Dict[str, t.List[str] | t.Dict[str, str]]]) -> t.Optional[t.Dict[str, t.List[str] | t.Dict[str, str]]]:
        """
        Validate the attributes field to ensure it is a dictionary.
        """
        if v is None: return {}
        for key in {'userdata', 'appdata', 'sparkdata'}:
            if v.get(key):
                try:
                    v[key] = json.loads(v[key][0] if isinstance(v[key], list) else v[key])
                except json.JSONDecodeError as e:
                    logger.error(f"Invalid JSON format for {key} in attributes: {e}")
                    raise HTTPException(status_code=400, detail=f"Invalid JSON format for {key} in attributes.") from e
        return v


    @field_serializer("attributes", when_used='json')
    def serialize_attributes(self, v: t.Optional[t.Dict[str, t.List[str] | t.Dict[str, str]]]) -> t.Optional[t.Dict[str, t.List[str] | t.Dict[str, str]]]:
        """
        Serialize the attributes field to ensure it is a dictionary.
        """
        if v is None: return {}
        for key in {'userdata', 'appdata', 'sparkdata', 'orgdata', 'litellmdata', 'kcdata'}:
            if isinstance(v.get(key), dict):
                v[key] = [json.dumps(v[key])]
            elif isinstance(v.get(key), list):
                v[key] = [json.dumps(item) for item in v[key] if isinstance(item, dict)]
        return v

    @model_validator(mode='after')
    def validate_user(self, info: ValidationInfo) -> KeycloakUser:
        """
        Validate the Keycloak user model after initial validation.
        """
        if info.context and 'realm' in info.context:
            self._extra['realm'] = info.context['realm']
        return self

    @eproperty
    def created_at(self) -> t.Optional[datetime.datetime]:
        """
        Returns the creation timestamp as a datetime object.

        >>> 1755145228090 -> datetime.datetime(2024, 5, 12, 10, 20, 28, 90)
        """
        if self.created_timestamp is None: return None
        return datetime.datetime.fromtimestamp(self.created_timestamp / 1000.0)

    @eproperty
    def appdata(self) -> t.Dict[str, t.Any]:
        """
        Returns the appdata attribute as a dictionary.
        """
        if 'appdata' not in self.attributes:
            self.attributes['appdata'] = {}
        return self.attributes['appdata']
    
    @eproperty
    def userdata(self) -> t.Dict[str, t.Any]:
        """
        Returns the userdata attribute as a dictionary.
        """
        if 'userdata' not in self.attributes:
            self.attributes['userdata'] = {}
        return self.attributes['userdata']

    @eproperty
    def sparkdata(self) -> t.Dict[str, t.Any]:
        """
        Returns the sparkdata attribute as a dictionary.
        """
        if 'sparkdata' not in self.attributes:
            self.attributes['sparkdata'] = {}
        return self.attributes['sparkdata']
    
    @eproperty
    def orgdata(self) -> t.Dict[str, t.Any]:
        """
        Returns the orgdata attribute as a dictionary.
        """
        if 'orgdata' not in self.attributes:
            self.attributes['orgdata'] = {}
        return self.attributes['orgdata']
    
    @eproperty
    def litellmdata(self) -> t.Dict[str, t.Any]:
        """
        Returns the litellmdata attribute as a dictionary.
        """
        if 'litellmdata' not in self.attributes:
            self.attributes['litellmdata'] = {}
        return self.attributes['litellmdata']
    
    @eproperty
    def kcdata(self) -> t.Dict[str, t.Any]:
        """
        Returns the kcdata attribute as a dictionary.
        """
        if 'kcdata' not in self.attributes:
            self.attributes['kcdata'] = {}
        return self.attributes['kcdata']

    @eproperty
    def realm(self) -> str:
        """
        Returns the realm of the user.
        """
        return self._extra.get('realm')


class KeycloakGroup(BaseKCModel):
    """
    Represents a Keycloak group.
    """
    id: str
    name: str
    description: t.Optional[str] = None
    path: str
    sub_groups: t.List[KeycloakGroup] = Field(default_factory=list)
    # access: t.Optional[t.Dict[str, t.Any]] = None

    @model_validator(mode='after')
    def validate_group(self, info: ValidationInfo) -> KeycloakGroup:
        """
        Validate the Keycloak group model after initial validation.
        """
        if info.context and 'realm' in info.context:
            self._extra['realm'] = info.context['realm']
        return self


class KeycloakWebhookCreate(BaseKCModel):
    """ 
    Represents a Keycloak webhook creation request.
    """
    url: str = Field(..., description="The URL to which the webhook will send events")
    enabled: bool = Field(default=True, description="Whether the webhook is enabled")
    secret: t.Optional[str] = Field(None, description="The secret for the webhook")
    event_types: t.List[str] = Field(default=["*"], description="List of event types for the webhook")


class KeycloakWebhookObject(BaseKCModel):
    """
    Represents a Keycloak webhook object.
    """
    id: t.Optional[str] = Field(None, description="The ID of the webhook object")
    enabled: bool = Field(default=True, description="Whether the webhook object is enabled")
    url: str = Field(..., description="The URL of the webhook object")
    secret: t.Optional[str] = Field(None, description="The secret for the webhook object")
    event_types: t.List[str] = Field(default_factory=list, description="List of event types for the webhook object")
    created_by: t.Optional[str] = Field(None, description="The user id who created the webhook object")
    created_at: t.Optional[datetime.datetime] = Field(None, description="The creation date of the webhook object")


class AuthDetails(BaseKCModel):
    realm_id: str = Field(..., description="The ID of the realm")
    client_id: str = Field(..., description="The ID of the client")
    user_id: str = Field(..., description="The ID of the user")
    ip_address: str = Field(..., description="The IP address of the user")
    username: str = Field(..., description="The username of the user")
    session_id: str = Field(..., description="The ID of the session")

# Valid Schema for access.CODE_TO_TOKEN, access.LOGIN

class AccessWebhookEvent(BaseKCModel):
    id: str = Field(..., description="The ID of the webhook event")
    time: int = Field(..., description="The time of the webhook event")
    realm_id: str = Field(..., description="The ID of the realm")
    realm_name: str = Field(..., description="The name of the realm")
    uid: str = Field(..., description="The user ID")
    auth_details: AuthDetails = Field(..., description="The authentication details")
    details: t.Dict[str, t.Any] = Field(..., description="The event details")
    type: str = Field(..., description="The event type")


from __future__ import annotations

"""
Base OpenAI Types
"""
import os
import json
import datetime
import pathlib
import tempfile
from enum import Enum
from lzo.types.base import BaseModel, Field, eproperty
from lzl.types.base import get_pydantic_field_names, ByteSize, validator, get_pydantic_fields
from lzl.api.openai.version import DEFAULT_AZURE_VERSION
from lzo.utils.serialization.serialization import ObjectDecoder, ObjectEncoder
from ..utils.helpers import aparse_stream, parse_stream
from ..utils.logs import logger
from typing import Optional, Union, Dict, List, Any, Type, Tuple, Iterator, AsyncIterator, TYPE_CHECKING


try:
    from lzl.io import File, FileLike as FileType
    _has_fileio = True
except ImportError:
    from pathlib import Path as File
    FileType = Union[File, str, os.PathLike]
    _has_fileio = False


if TYPE_CHECKING:
    from lzl.api.aiohttpx import Response

class ApiType(str, Enum):
    openai = "openai"
    azure = "azure"
    azure_ad = "azure_ad"

    def get_version(
        self, 
        version: Optional[str] = None
    ):
        """
        Returns the API Version
        """
        if self.value in {"azure", "azure_ad", "azuread"} and not version:
            return DEFAULT_AZURE_VERSION
        return version

    @classmethod
    def parse(cls, value: str) -> "ApiType":
        """
        Parses the API Type
        """
        return cls(value) if "azure" in value.lower() else cls.openai




class FilePurpose(str, Enum):
    """
    File Purpose
    """

    finetune = "fine-tune"
    fine_tune = "fine-tune"
    train = "fine-tune-train"
    search = "search"
    batch = "batch"

    @classmethod
    def parse_str(cls, value: Union[str, 'FilePurpose'], raise_error: bool = True):
        if isinstance(value, cls): return value
        if "train" in value:
            return cls.train
        elif "finetune" in value:
            return cls.finetune
        elif "fine-tune" in value:
            return cls.fine_tune
        elif "search" in value:
            return cls.search
        elif "batch" in value:
            return cls.batch
        if not raise_error: return None
        raise ValueError(f"Cannot convert {value} to FilePurpose")

class Usage(BaseModel):
    prompt_tokens: Optional[int] = 0
    completion_tokens: Optional[int] = 0
    total_tokens: Optional[int] = 0

    # @lazyproperty
    @property
    def consumption(self) -> int:
        """
        Gets the consumption
        """
        return self.total_tokens
    
    def update(self, usage: Union['Usage', Dict[str, int]]):
        """
        Updates the consumption
        """
        if isinstance(usage, Usage):
            if usage.prompt_tokens: self.prompt_tokens += usage.prompt_tokens
            if usage.completion_tokens: self.completion_tokens += usage.completion_tokens
            if usage.total_tokens: self.total_tokens += usage.total_tokens
            return
        if usage.get('prompt_tokens'): self.prompt_tokens += usage['prompt_tokens']
        if usage.get('completion_tokens'): self.completion_tokens += usage['completion_tokens']
        if usage.get('total_tokens'): self.total_tokens += usage['total_tokens']

    def __iadd__(self, other: Union['Usage', Dict[str, int]]):
        """
        Adds the usage
        """
        self.update(other)
        return self.consumption


class ModelCosts(BaseModel):
    """
    Represents a model's costs
    """
    unit: Optional[int] = 1000
    input: Optional[float] = 0.0
    output: Optional[float] = 0.0
    total: Optional[float] = 0.0


class ModelCostItem(BaseModel):
    """
    Represents a model's Cost Item
    """
    name: str
    aliases: Optional[List[str]] = None
    context_length: Optional[int] = 0
    costs: Optional[ModelCosts] = Field(default_factory=ModelCosts)
    batch_costs: Optional[ModelCosts] = None
    endpoints: Optional[List[str]] = None

    # Adding new properties
    data: Optional[Dict[str, Any]] = Field(default_factory=dict)

    def get_costs(
        self, 
        input_tokens: Optional[int] = None,
        output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        usage: Optional[Union['Usage', Dict[str, int]]] = None,
        is_batch: Optional[bool] = None,
        **kwargs
    ) -> float:
        """
        Gets the costs
        """
        if usage is not None:
            if isinstance(usage, dict): 
                usage = Usage(**usage)
            input_tokens = usage.prompt_tokens
            output_tokens = usage.completion_tokens
        if kwargs.get('prompt_tokens'): input_tokens = kwargs['prompt_tokens']
        if kwargs.get('completion_tokens'): output_tokens = kwargs['completion_tokens']
        assert input_tokens is not None or output_tokens is not None or total_tokens is not None, "Must provide either input_tokens, output_tokens, or total_tokens"
        cost_ref = self.batch_costs if is_batch else self.costs
        if cost_ref is None: return 0.0
        cost = 0.0
        if cost_ref.input: cost += cost_ref.input * input_tokens / cost_ref.unit
        if cost_ref.output: cost += cost_ref.output * output_tokens / cost_ref.unit
        if cost_ref.total and total_tokens is not None: cost += cost_ref.total * total_tokens / cost_ref.unit
        return cost

    def supports(self, endpoint: str) -> bool:
        """
        Returns True if the endpoint is supported
        """
        return endpoint in self.endpoints if self.endpoints else False


class BaseResource(BaseModel):

    """
    Base Object class for resources to
    inherit from
    """


    if TYPE_CHECKING:
        id: Optional[str]
        file_id: Optional[str]
        fine_tune_id: Optional[str]
        model_id: Optional[str]
        completion_id: Optional[str]
        openai_id: Optional[str]
        model: Optional[str]


    @eproperty
    def resource_id(self):
        """
        Returns the resource id
        """
        if hasattr(self, 'id'):
            return self.id
        if hasattr(self, 'file_id'):
            return self.file_id
        if hasattr(self, 'fine_tune_id'):
            return self.fine_tune_id
        if hasattr(self, 'model_id'):
            return self.model_id
        if hasattr(self, 'completion_id'):
            return self.completion_id
        return self.openai_id if hasattr(self, 'openai_id') else None
    
    @staticmethod
    def create_resource(
        resource: Type['BaseResource'],
        **kwargs
    ) -> Tuple['BaseResource', Dict]:
        """
        Extracts the resource from the kwargs and returns the resource 
        and the remaining kwargs
        """
        resource_fields = get_pydantic_field_names(resource)
        resource_kwargs = {k: v for k, v in kwargs.items() if k in resource_fields}
        return_kwargs = {k: v for k, v in kwargs.items() if k not in resource_fields}
        resource_obj = resource.model_validate(resource_kwargs)
        return resource_obj, return_kwargs
    

    @staticmethod
    def create_batch_resource(
        resource: Type['BaseResource'],
        batch: List[Union[Dict[str, Any], Any]],
        **kwargs
    ) -> Tuple[List['BaseResource'], Dict]:
        """
        Extracts the resource from the kwargs and returns the resource 
        and the remaining kwargs
        """
        resource_fields = get_pydantic_field_names(resource)
        resource_kwargs = {k: v for k, v in kwargs.items() if k in resource_fields}
        return_kwargs = {k: v for k, v in kwargs.items() if k not in resource_fields}
        resource_objs = []
        for item in batch:
            if isinstance(item, dict):
                item.update(resource_kwargs)
                resource_objs.append(resource.model_validate(item))
            else:
                resource_objs.append(item)
        return resource_objs, return_kwargs

    @classmethod
    def create_many(cls, data: List[Dict]) -> List['BaseResource']:
        """
        Creates many resources
        """
        return [cls.model_validate(d) for d in data]
    
    @staticmethod
    def handle_json(
        content: Any,
        **kwargs
    ) -> Union[Dict, List]:
        """
        Handles the json response
        """
        return json.loads(content, cls = ObjectDecoder, **kwargs)


    @staticmethod
    def handle_stream(
        response: 'Response',
        streaming: Optional[bool] = False,
    ) -> Iterator[Dict]:
        """
        Handles the stream response
        """
        for line in parse_stream(response):
            if not line.strip(): continue
            try:
                yield json.loads(line)
            except Exception as e:
                logger.error(f'Error: {line}: {e}')
    
    @staticmethod
    async def ahandle_stream(
        response: 'Response',
        streaming: Optional[bool] = False,
    ) -> AsyncIterator[Dict]:
        """
        Handles the stream response
        """
        async for line in aparse_stream(response):
            if not line.strip(): continue
            try:
                yield json.loads(line)
            except Exception as e:
                logger.error(f'Error: {line}: {e}')


    def __getitem__(self, key: str) -> Any:
        """
        Mimic dict
        """
        return getattr(self, key)


class Permission(BaseResource):
    id: str
    object: str
    created: datetime.datetime
    allow_create_engine: bool
    allow_sampling: bool
    allow_logprobs: bool
    allow_search_indices: bool
    allow_view: bool
    allow_fine_tuning: bool
    organization: str
    group: Optional[str]
    is_blocking: bool

    @property
    def since_seconds(self):
        return (datetime.datetime.now(datetime.timezone.utc) - self.created).total_seconds()


class FileObject(BaseResource):
    id: str
    object: Optional[str] = 'file'
    bytes: Optional[ByteSize]
    created_at: Optional[datetime.datetime]
    filename: Optional[str]
    purpose: Optional[FilePurpose] = FilePurpose.fine_tune

    @validator("created_at")
    def validate_created_at(cls, value):
        return datetime.datetime.fromtimestamp(value, datetime.timezone.utc) if value else value
    
    @classmethod
    def create_many(cls, data: List[Dict]) -> List['FileObject']:
        """
        Creates many resources
        """
        return [cls.model_validate(d) for d in data]

class EventObject(BaseResource):
    object: Optional[str]
    created_at: Optional[datetime.datetime]
    level: Optional[str]
    message: Optional[str]

    @property
    def since_seconds(self) -> int:
        if self.created_at is None: return -1
        return (datetime.datetime.now(datetime.timezone.utc) - self.created_at).total_seconds()


class FileResource(BaseResource):
    file: Optional[Union[str, FileType, Any]] # type: ignore
    file_id: Optional[str]
    filename: Optional[str] = None
    purpose: FilePurpose = FilePurpose.fine_tune
    model: Optional[str] = None

    @validator("purpose")
    def validate_purpose(cls, value):
        return FilePurpose.parse_str(value) if isinstance(value, str) else value
    
    def get_params(self, **kwargs) -> List:
        """
        Transforms the data to the req params
        """
        files = [("purpose", (None, self.purpose.value))]
        if self.purpose == FilePurpose.search and self.model:
            files.append(("model", (None, self.model)))
        if self.file:
            file = File(self.file)
            files.append(
                ("file", (self.filename or file.name, file.read_bytes(), "application/octet-stream"))
            )
        return files
    
    async def async_get_params(self, **kwargs) -> List:
        """
        Transforms the data to the req params
        """
        files = [("purpose", (None, self.purpose.value))]
        if self.purpose == FilePurpose.search and self.model:
            files.append(("model", (None, self.model)))
        if self.file:
            file = File(self.file)
            files.append(
                ("file", (self.filename or file.name, (await file.async_read_bytes() if _has_fileio else file.read_bytes()), "application/octet-stream"))
            )
        return files

    @classmethod
    def create_from_batch(
        cls,
        batch: List[Union[Dict[str, Any], str]],
        output_path: Optional[str] = None,
        file_id: Optional[str] = None,
        filename: Optional[str] = None,
        purpose: Optional[FilePurpose] = None,
        **kwargs,
    ) -> Tuple['FileObject', Dict[str, Any]]:
        """
        Creates a file object from a batch in jsonl format
        """
        for n, b in enumerate(batch):
            if isinstance(b, dict):
                batch[n] = json.dumps(b, cls = ObjectEncoder)
        if output_path:
            output = pathlib.Path(output_path)
        else:
            tmp = tempfile.NamedTemporaryFile(delete = False)
            tmp.close()
            output = pathlib.Path(tmp.name)

        with output.open('w') as f:
            for b in batch:
                f.write(f'{b}\n')
        resource_fields = get_pydantic_field_names(cls)
        resource_kwargs = {k: v for k, v in kwargs.items() if k in resource_fields}
        return_kwargs = {k: v for k, v in kwargs.items() if k not in resource_fields}
        return cls(
            file = output,
            purpose = purpose,
            filename = filename,
            file_id = file_id,
            **resource_kwargs
        ), return_kwargs
        



"""
Fork of `async_openai` to continue extending the library
"""

from typing import TYPE_CHECKING

from .assets import (
    load_provider_prices,
    load_preset_config,
)
from .configs import (
    OpenAISettings, 
    AzureOpenAISettings,
)

from .types import (
    Usage,
    OpenAIError,
    RateLimitError,
    MaxRetriesExhausted,
    MaxRetriesExceeded,
)

from .clients import (
    OpenAI,
    OpenAIClient,
    OpenAIManager,
    OpenAIFunctions,
    FunctionManager as OpenAIFunctionsManager,
)

if TYPE_CHECKING:
    from .schemas import (
        ChatMessage,
        ChatChoice,
        ChatResponse,
        ChatObject,

        CompletionChoice,
        CompletionObject,
        CompletionResponse,

        EmbeddingData,
        EmbeddingObject,
        EmbeddingResponse,

        BaseFunctionModel, 
        BaseFunction,
    )


from __future__ import annotations

"""
Modal Classes and Objects that serve as the base class that can be subclassed
"""
import os
import sys
import contextlib
import typing as t
from lzl import load

if load.TYPE_CHECKING:
    import torch
else:
    torch = load.LazyLoad("torch", install_missing=True)


class NoStdStreams(object):
    def __init__(self):
        self.devnull = open(os.devnull, "w")

    def __enter__(self):
        self._stdout, self._stderr = sys.stdout, sys.stderr
        self._stdout.flush(), self._stderr.flush()
        sys.stdout, sys.stderr = self.devnull, self.devnull

    def __exit__(self, exc_type, exc_value, traceback):
        sys.stdout, sys.stderr = self._stdout, self._stderr
        self.devnull.close()


class ModalClass:
    """
    Base class for Modal classes.
    """

    def _setup_(self, **kwargs):
        """
        Setup method that initializes certain properties or configurations.
        """
        from lzl.logging import logger
        from lzo.utils import Timer
        from .static import MODAL_GPU_PRICING

        self.t = Timer()
        self.timer = Timer
        self.logger = logger

        self.device_name = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(self.device_name)
        self.has_gpu = torch.cuda.is_available()
        self.gpu_device_name: t.Optional[str] = None
        self.gpu_cost_per_sec: float = 0.0
        if self.has_gpu:
            # NVIDIA A2
            from lzo.utils.system import get_gpu_data
            gpu_data = get_gpu_data()
            self.gpu_device_name = gpu_data['name'] if gpu_data else None
            self.gpu_cost_per_sec = MODAL_GPU_PRICING.get(self.gpu_device_name, 0.0)
            self.logger.info(f"Using GPU: |g|{self.gpu_device_name}|e|", colored = True, prefix = self.model_name)
        
        self.cold_start_time = self.t.total
        self.cold_start_time_str = self.timer.pformat_duration(self.cold_start_time)
        self.last_req_duration: float = 0.0
        self.last_req_batch_size: int = 0
        self.last_req_cost: float = 0.0
        self.last_gpu_memory: t.Optional[str] = None
        self._load_model_(**kwargs)
        self.logger.info(f"Loading Model: {self.model_id} in |g|{self.t.total_s}|e| on {self.device_name}", colored = True, prefix = self.model_name)
        self.num_requests: int = 0
        self.batches_handled: int = 0
        self.total_duration: float = 0.0


    def _load_model_(self, **kwargs):
        """
        Load the model with the given keyword arguments.
        """
        pass
    
    @property
    def model_id(self) -> t.Optional[str]:
        """
        Returns the model ID.
        """
        return getattr(self, 'MODEL_ID', None)
    
    @property
    def model_name(self) -> t.Optional[str]:
        """
        Returns the model name.
        """
        return getattr(self, 'MODEL_NAME', None)
    
    def update_metadata(self, metadata: t.Dict[str, t.Any]) -> None:
        """
        Updates the metadata
        """
        metadata.update({
            'cost_value': self.last_req_cost,
            'cost_string': f"${self.last_req_cost:,.5f}",
            'batch_size': self.last_req_batch_size,
            'duration_value': self.last_req_duration,
            'duration_string': self.timer.pformat_duration(self.last_req_duration, include_ms = True, short = 1),
            'cold_start_time_string': self.cold_start_time_str,
            'cold_start_time_value': self.cold_start_time,
        })
        if self.has_gpu: metadata['gpu_memory'] = self.last_gpu_memory
    
    """
    Utility functions
    """

    def _prestop_(self):
        """
        The exit point for the model
        """
        self.logger.info(f"|r|[Shutting Down]|e| Completed Requests: |g|{self.num_requests}|e|. Total Batches Handled: |g|{self.batches_handled}|e|", colored = True, prefix = self.model_name)
        self.logger.info(f"|r|[Shutting Down]|e| Total Inference Duration: |g|{self.timer.pformat_duration(self.total_duration)}|e|. Total Time Alive: |g|{self.t.total_s}|e|", colored = True, prefix = self.model_name)
        if self.gpu_cost_per_sec:
            self.logger.info(f"|r|[Shutting Down]|e| Total Cost: |y|${self.t.total * self.gpu_cost_per_sec:,.4f}|e|", colored = True, prefix = self.model_name)
        if self.has_gpu:
            self.logger.info(f"|r|[Shutting Down]|e| {self.get_gpu_memory_nvidia_smi(colored = True)}", colored = True, prefix = self.model_name)


    @contextlib.contextmanager
    def default_mode(self):
        """
        Handles the default mode
        """
        t = self.timer(format_ms = True, format_short = 1)
        try:
            yield
        except Exception as e:
            self.logger.error(f"[{self.model_id}] Error in default mode: {e}")
            raise e
        finally:
            total_s = t.total
            self.last_req_duration = total_s
            self.total_duration += total_s
            self.num_requests += 1
            if self.gpu_cost_per_sec:
                self.last_req_cost = total_s * self.gpu_cost_per_sec
            if self.has_gpu:
                self.last_gpu_memory = self.get_gpu_memory_nvidia_smi(colored = False)
            self.logger.info(f"Total Requests: |g|{self.num_requests}|e|. Total Batches Handled: |g|{self.batches_handled}|e|. Handled Last Batch Size of |g|{self.last_req_batch_size}|e| in |g|{t.total_s}|e|", colored = True, prefix = self.model_name)
            self.logger.info(f"Total Inference Duration: |g|{self.timer.pformat_duration(self.total_duration)}|e|. Total Time Alive: |g|{self.t.total_s}|e|", colored = True, prefix = self.model_name)
            self.logger.info(f"Total Cost: |y|${self.last_req_cost:,.5f}|e|", colored = True, prefix = self.model_name)
            if self.has_gpu:
                self.logger.info(f"{self.get_gpu_memory_nvidia_smi(colored = True)}", colored = True, prefix = self.model_name)

    
    @contextlib.contextmanager
    def inference_mode(self):
        """
        Handles the inference mode
        """
        with torch.inference_mode():
            t = self.timer(format_ms = True, format_short = 1)
            try:
                yield
            except Exception as e:
                self.logger.error(f"[{self.model_id}] Error in inference mode: {e}")
                raise e
            finally:
                total_s = t.total
                self.last_req_duration = total_s
                self.total_duration += total_s
                self.num_requests += 1
                if self.gpu_cost_per_sec:
                    self.last_req_cost = total_s * self.gpu_cost_per_sec
                if self.has_gpu:
                    self.last_gpu_memory = self.get_gpu_memory_nvidia_smi(colored = False)
                self.logger.info(f"Total Requests: |g|{self.num_requests}|e|. Total Batches Handled: |g|{self.batches_handled}|e|. Handled Last Batch Size of |g|{self.last_req_batch_size}|e| in |g|{t.total_s}|e|", colored = True, prefix = self.model_name)
                self.logger.info(f"Total Inference Duration: |g|{self.timer.pformat_duration(self.total_duration)}|e|. Total Time Alive: |g|{self.t.total_s}|e|", colored = True, prefix = self.model_name)
                self.logger.info(f"Total Cost: |y|${self.last_req_cost:,.5f}|e|", colored = True, prefix = self.model_name)
                if self.has_gpu:
                    self.logger.info(f"{self.get_gpu_memory_nvidia_smi(colored = True)}", colored = True, prefix = self.model_name)


    def get_gpu_memory_nvidia_smi(self, short: bool = None, colored: bool = False) -> t.Optional[str]:
        """
        Returns the GPU memory
        """
        from lzo.utils.system import get_gpu_data
        gpu_data = get_gpu_data()
        if not gpu_data: return None
        mem_used = gpu_data['memory_used']
        if short: return mem_used.human_readable()
        mem_percent = gpu_data['utilization_memory']
        mem_total = gpu_data['memory_total']
        gpu_name = gpu_data['name']
        if not colored: return f"{gpu_name}: {mem_used.human_readable()} / {mem_total.human_readable()} ({mem_percent}%)"
        return f"{gpu_name}: |g|{mem_used.human_readable()}|e| / |y|{mem_total.human_readable()}|e| ({mem_percent}%)"
    


"""
Because FSSpec is inconsistent in name mapping, we gotta do this.
"""

ASYNC_METHODS = [
    "acall",
    "acat",
    "acat_file",
    "acat_ranges",
    "acopy",
    "acp_file",
    "adu",
    "aexists",
    "aexpand_path",
    "afind",
    "aget",
    "aget_file",
    "agetxattr",
    "aglob",
    "ainfo",
    "ametadata",
    "aisdir",
    "aisfile",
    "alist_buckets",
    "alist_objects",
    "als",
    "amakedirs",
    "amerge",
    "amkdir",
    "apipe",
    "apipe_file",
    "aprocess_limits",
    "aput",
    "aput_file",
    "arm",
    "arm_file",
    "armdir",
    "arm_files",
    "asetxattrs",
    "asize",
    "asizes",
]

ASYNC_MAPPING_GS = {
    "_call": "acall",
    "_cat": "acat",
    "_cat_file": "acat_file",
    "_cat_ranges": "acat_ranges",
    "_copy": "acopy",
    "_cp_file": "acp_file",
    "_du": "adu",
    "_exists": "aexists",
    "_expand_path": "aexpand_path",
    "_find": "afind",
    "_get": "aget",
    "_get_file": "aget_file",
    "_getxattr": "agetxattr",
    "_glob": "aglob",
    "_info": "ainfo",
    "_metadata": "ametadata",
    "_isdir": "aisdir",
    "_isfile": "aisfile",
    "_list_buckets": "alist_buckets",
    "_list_objects": "alist_objects",
    "_ls": "als",
    #"_makedir": "amakedir",
    "_makedirs": "amakedirs",
    "_merge": "amerge",
    "_mkdir": "amkdir",
    "_pipe": "apipe",
    "_pipe_file": "apipe_file",
    "_process_limits": "aprocess_limits",
    "_put": "aput",
    "_put_file": "aput_file",
    "_rm": "arm",
    "_rm_file": "arm_file",
    "_rmdir": "armdir",
    "_rm_files": "arm_files",
    "_setxattrs": "asetxattrs",
    "_size": "asize",
    "_sizes": "asizes",
}


ASYNC_MAPPING_S3 = {
    "_call_s3": "acall",
    "_cat": "acat",
    "_cat_file": "acat_file",
    "_cat_ranges": "acat_ranges",
    "_copy": "acopy",
    "_cp_file": "acp_file",
    "_du": "adu",
    "_exists": "aexists",
    "_expand_path": "aexpand_path",
    "_find": "afind",
    "_get": "aget",
    "_get_file": "aget_file",
    "_getxattr": "agetxattr",
    "_glob": "aglob",
    "_info": "ainfo",
    "_metadata": "ametadata",
    "_isdir": "aisdir",
    "_isfile": "aisfile",
    "_lsbuckets": "alist_buckets",
    "_lsdir": "alist_objects",
    "_ls": "als",
    "_makedirs": "amakedirs",
    "_merge": "amerge",
    "_mkdir": "amkdir",
    "_pipe": "apipe",
    "_pipe_file": "apipe_file",
    "_process_limits": "aprocess_limits",
    "_put": "aput",
    "_put_file": "aput_file",
    "_rm": "arm",
    "_rm_file": "arm_file",
    "_rmdir": "armdir",
    "_bulk_delete": "arm_files",
    "_setxattr": "asetxattrs",
    "_size": "asize",
    "_sizes": "asizes",
}


ASYNC_MAPPING_AZ = {
    "_call": "acall",
    "_cat": "acat",
    "_cat_file": "acat_file",
    "_cat_ranges": "acat_ranges",
    "_copy": "acopy",
    "_cp_file": "acp_file",
    "_du": "adu",
    "_exists": "aexists",
    "_expand_path": "aexpand_path",
    "_find": "afind",
    "_get": "aget",
    "_get_file": "aget_file",
    "_getxattr": "agetxattr",
    "_glob": "aglob",
    "_info": "ainfo",
    "_metadata": "ametadata",
    "_isdir": "aisdir",
    "_isfile": "aisfile",
    "_list_buckets": "alist_buckets",
    "_list_objects": "alist_objects",
    "_ls": "als",
    #"_makedir": "amakedir",
    "_makedirs": "amakedirs",
    "_merge": "amerge",
    "_mkdir": "amkdir",
    "_pipe": "apipe",
    "_pipe_file": "apipe_file",
    "_process_limits": "aprocess_limits",
    "_put": "aput",
    "_put_file": "aput_file",
    "_rm": "arm",
    "_rm_file": "arm_file",
    "_rmdir": "armdir",
    "_rm_files": "arm_files",
    "_setxattrs": "asetxattrs",
    "_size": "asize",
    "_sizes": "asizes",
}

ASYNC_SYNTAX_MAPPING = {
    'gs': ASYNC_MAPPING_GS,
    's3': ASYNC_MAPPING_S3,
    'minio': ASYNC_MAPPING_S3,
    'az': ASYNC_MAPPING_AZ,
}


from __future__ import annotations

"""
Kubernetes Client based on kr8s
"""
import os
import abc
import typing as t
from pathlib import Path
from lzl import load
from lzl.proxied import ProxyObject
from lzl.logging import logger

if load.TYPE_CHECKING:
    import kr8s
    import httpx
else:
    kr8s = load.LazyLoad("kr8s", install_missing=True)
    httpx = load.LazyLoad("httpx", install_missing=True)

from .context import KubernetesContext
    

class KubernetesClient(abc.ABC):
    """
    The Kubernetes Client
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the Kubernetes Client

        - For use in Kubernetes, the client will automatically use the in-cluster configuration
        - For use outside of Kubernetes, the client will use the kubeconfigs directory
          env: KUBECONFIGS_DIR
        """
        from lzo.utils.system import is_in_kubernetes
        self.in_k8s = is_in_kubernetes()
        self.kconfigs_path = None
        if not self.in_k8s:
            if kctx_dir := os.getenv('KUBECONFIGS_DIR', os.getenv('KUBECONFIG_DIR', os.path.expanduser('~/.kube'))):
                self.kconfigs_path = Path(kctx_dir)
        self._ctxs: t.Dict[str, KubernetesContext] = {}
        self._default: t.Optional[str] = None
        self._kwargs = kwargs

    
    def find_kubeconfig_ctx(self, name: str) -> t.Optional[Path]:
        """
        Find a kubeconfig context
        """
        if self.kconfigs_path:
            for ctx_fname in self.kconfigs_path.iterdir():
                if name in ctx_fname.name:
                    logger.info(f'Using kubeconfig: `{ctx_fname.name}`')
                    return ctx_fname
        return None

    def get_or_init(
        self,
        name: t.Optional[str] = None, 
        url: t.Optional[str] = None,
        kubeconfig: t.Optional[str] = None,
        service_account: t.Optional[str] = None,
        context: t.Optional[str] = None,
        namespace: t.Optional[str] = None,
        set_as_default: t.Optional[bool] = False,
        **kwargs
    ) -> KubernetesContext:
        """
        Get or initialize a context
        """
        name = name if name is not None else (self._default or 'default')
        if name not in self._ctxs:
            if name != 'default' and not kubeconfig:
                kubeconfig = self.find_kubeconfig_ctx(name)
            self._ctxs[name] = KubernetesContext(
                name = name,
                url = url,
                kubeconfig = kubeconfig,
                service_account = service_account,
                context = context,
                namespace = namespace,
                **kwargs
            )
            if self._default is None or set_as_default:
                self._default = name
        return self._ctxs[name]

    @property
    def ctx(self) -> KubernetesContext:
        """
        Returns the default context
        """
        if self._default is None:
            return self.get_or_init()
        return self._ctxs[self._default]
    
    async def ainit(self) -> None:
        """
        Initialize the client
        """
        if not self._ctxs:
            await self.ctx.ainit()
        else:
            for ctx in self._ctxs.values():
                await ctx.ainit()

    def __getitem__(self, name: str) -> KubernetesContext:
        """
        Get a context
        """
        return self._ctxs[name]
    

Kubernetes: KubernetesClient = ProxyObject(
    obj_cls = KubernetesClient,
)


from __future__ import annotations

from typing import Union, Optional, Dict, Any, List, TYPE_CHECKING

if TYPE_CHECKING:
    from fastapi import Request as BaseRequest
    from .user import OAuth2User
    from .token import OAuth2Token, AccessToken, IDToken


    class Request(BaseRequest):

        @property
        def auth(self) -> OAuth2Token:
            """
            Returns the OAuth2 Token
            """
            ...


        @property
        def user(self) -> OAuth2User:
            """
            Returns the OAuth2 User
            """
            ...






"""
This library maintains its own versioning
"""

VERSION = '0.0.1'

DEFAULT_AZURE_VERSION = '2024-12-01-preview'
DEFAULT_FUNCTION_MODEL = 'gpt-4o-mini'
DEFAULT_CHAT_MODEL = 'gpt-4o-mini'
DEFAULT_COMPLETION_MODEL = 'gpt-4o-mini'
DEFAULT_EMBEDDING_MODEL = 'text-embedding-3-small'


from __future__ import annotations

"""Key generation helpers used for secrets, tokens, and passcodes."""

import os
import sys
import hashlib
import string
import secrets
import base64
import typing as t
import uuid as _uuid
from lzl import load

if t.TYPE_CHECKING:
    import bcrypt
else:
    bcrypt = load.lazy_load("bcrypt")


ALPHA_NUMERIC = string.ascii_letters + string.digits

__all__ = [
    'ALPHA_NUMERIC',
    'Base64',
    'Generate',
    'generate_htpasswd_key',
    'validate_htpasswd_key',
]

class Base64:
    """Convenience wrapper for base64 encoding/decoding strings."""

    encoding: str = 'utf-8'

    @classmethod
    def encode(cls, text: str, encoding: str = 'utf-8', *args: t.Any, **kwargs: t.Any) -> str:
        """Encode ``text`` using ``base64.b64encode``."""

        return base64.b64encode(text.encode(encoding=encoding), *args, **kwargs).decode(encoding=encoding)

    @classmethod
    def decode(
        cls,
        data: t.Union[str, bytes],
        encoding: str = 'utf-8',
        *args: t.Any,
        **kwargs: t.Any,
    ) -> str:
        """Decode base64 data back into a UTF-8 string."""

        if isinstance(data, str):
            data = data.encode(encoding=encoding)
        return base64.b64decode(data, *args, **kwargs).decode(encoding=encoding)

    @classmethod
    def dumps(cls, data: str, encoding: str = 'utf-8', *args: t.Any, **kwargs: t.Any) -> str:
        """Alias for :meth:`encode` to mirror JSON-like APIs."""

        return cls.encode(data, encoding=encoding, *args, **kwargs)

    @classmethod
    def loads(
        cls,
        data: t.Union[str, bytes],
        encoding: str = 'utf-8',
        *args: t.Any,
        **kwargs: t.Any,
    ) -> str:
        """Alias for :meth:`decode` to mirror JSON-like APIs."""

        return cls.decode(data, encoding=encoding, *args, **kwargs)


class Generate:
    """Helpers for creating pseudo-random identifiers and secrets."""

    default_method: str = 'uuid4'

    @classmethod
    def uuid(cls, method: t.Optional[str] = None, *args: t.Any, **kwargs: t.Any) -> str:
        """Return a UUID string using ``method`` (falls back to ``uuid4``)."""

        method_name = method or cls.default_method
        generator = getattr(_uuid, method_name, getattr(_uuid, cls.default_method))
        return str(generator(*args, **kwargs))

    @classmethod
    def uuid_passcode(
        cls,
        length: t.Optional[int] = None,
        clean: bool = True,
        method: t.Optional[str] = None,
        raw: bool = False,
    ) -> str:
        """Return a UUID-derived passcode with optional cleanup/truncation."""

        value = cls.uuid(method=method)
        if raw:
            return value
        if clean:
            value = value.replace('-', '').strip()
        if length:
            value = value[:length]
        return value

    @classmethod
    def alphanumeric_passcode(cls, length: int = 16, alpha_only: bool = False) -> str:
        """Return a random ASCII alphanumeric string of ``length`` characters."""

        select = string.ascii_letters if alpha_only else ALPHA_NUMERIC
        return ''.join(secrets.choice(select) for _ in range(length))

    @classmethod
    def token(cls, length: int = 32, safe: bool = False, clean: bool = True) -> str:
        """Generate URL-safe or hex tokens suitable for API keys."""

        value = secrets.token_hex(length) if safe else secrets.token_urlsafe(length)
        if clean:
            for char in value:
                if char not in ALPHA_NUMERIC:
                    value = value.replace(char, secrets.choice(ALPHA_NUMERIC))
        return value

    @classmethod
    def openssl_random_key(cls, length: int = 64, base: bool = True) -> str:
        """Mimic ``openssl rand`` with optional base64 encoding."""

        key = secrets.token_hex(length)
        if base:
            key = Base64.encode(key)
        return key

    @classmethod
    def keypair(cls, key_length: int = 16, secret_length: int = 36) -> t.Dict[str, str]:
        """Return a random key/secret pair using ``alphanumeric_passcode``."""

        return {
            'key': cls.alphanumeric_passcode(key_length),
            'secret': cls.alphanumeric_passcode(secret_length),
        }


def generate_htpasswd_key(
    secret: str,
    salt: t.Optional[str] = None,
    rounds: int = 10,
    repeat: int = 1,
) -> t.Generator[str, None, None]:
    """Yield bcrypt hashes suitable for ``htpasswd`` entries."""
    for _ in range(repeat):
        if salt:
            hashed = bcrypt.hashpw(secret.encode(), salt.encode())
        else:
            hashed = bcrypt.hashpw(secret.encode(), bcrypt.gensalt(rounds = rounds))
        yield hashed.decode()

def validate_htpasswd_key(
    secret: str,
    hashed: str,
) -> bool:
    """Return ``True`` when ``secret`` matches the bcrypt ``hashed`` value."""
    return bcrypt.checkpw(secret.encode(), hashed.encode())
    



import importlib

from typing import Any
from lzl.logging import logger, null_logger

logger.set_module_name('lzl.ext.temporal.loop', 'tmprl', is_relative = True)
logger.set_module_name('lzl.ext.temporal', 'temporal', is_relative = True)



from __future__ import annotations


import os
import typer
import subprocess
from pathlib import Path
from .static import COLOR
from typing import Optional, List, Any


def build_aliases(name: str, additional_names: Optional[List[str]] = None) -> List[str]:
    """
    Create the aliases for a given service name
    """
    aliases = [name]
    if '.' not in name:
        if '-' in name: aliases.append(name.replace('-', '.'))
        elif '_' in name: aliases.append(name.replace('_', '.'))
    if '-' not in name:
        if '.' in name: aliases.append(name.replace('.', '-'))
        elif '_' in name: aliases.append(name.replace('_', '-'))
    if '_' not in name:
        if '.' in name: aliases.append(name.replace('.', '_'))
        elif '-' in name: aliases.append(name.replace('-', '_'))
    if additional_names: aliases.extend(additional_names)
    return list(set(aliases))


def add_to_env(
    envvar: str,
    envval: Any,
    envpath: Optional[str] = '~/.bashrc',
):
    """
    Helper for adding to the environment variable
    """
    envval = str(envval)
    if ' ' in envval: envval = f'"{envval}"'
    os.system(f"echo 'export {envvar}={envval}' >> {envpath}")
    os.environ[envvar] = envval


def parse_text_file(path: Path) -> List[str]:
    """
    Parses a text file
    """
    text_lines = path.read_text().split('\n')
    return [line.strip() for line in text_lines if ('#' not in line[:5] and line.strip())]

def echo(message: str):
    """ 
    Helper for printing a message
    """
    typer.echo(message, color = True)

def run_cmd(cmdstr: str):
    """
    Helper for running a command
    """
    try:
        data = subprocess.check_output(cmdstr, shell=True, text=True, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
        if data[-1:] == '\n': data = data[:-1]
        echo(data)

    except subprocess.CalledProcessError as e:
        echo(f'{COLOR.RED}Failed to run command: {cmdstr}{COLOR.END}')
        echo(f'{COLOR.RED}Error: {e.stderr}{COLOR.END}')
        raise e




from __future__ import annotations

from .host import (
    get_host_name,
    get_host_ip,
    get_torch_device_name,
    get_torch_device,
    get_cpu_count,
    is_readonly_dir,
    fetch_resolver_nameserver,
    get_ulimits,
    set_ulimits,
)

from .kube import (
    get_k8s_namespace,
    get_k8s_kubeconfig,
    get_local_kubeconfig,
    get_local_kubeconfig_dir,
    is_in_kubernetes,
)
from .utils import (
    parse_memory_metric,
    parse_memory_metric_to_bs,
)
from .gpu import (
    get_gpu_data,
    aget_gpu_data,
)

from .resources import (
    get_resource_data
)